import sys
if __name__ == '__main__':
    sys.path.append('../')

from PyQt5.QtCore import Qt, QThread, QObject, pyqtSignal,QFileInfo
from PyQt5.QtWidgets import (QWidget, QApplication, QVBoxLayout, QTreeWidget,
                             QTreeWidgetItem, QTextEdit, QLineEdit, QPushButton,
                             QLabel, QHBoxLayout, QFileDialog,QTabWidget,
                             QFormLayout,QGridLayout,QListWidget,QSizePolicy,
                             QComboBox)
from PyQt5.QtGui import QFontMetrics,QFont

# Queue module version name changed between Python 2 & 3, so this allows
# for both
if sys.version[0] == '2':
    from Queue import Queue
else:
    from queue import Queue

import os,traceback,sys

import re

from cued_datalogger.api.channel import ChannelSet


class AddonManager(QWidget):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.parent = parent

        self.addon_functions = {}

        self.addon_local_vars = {}
        self.addon_global_vars = {}

        self.addon_writer = None
        self.init_ui()

        # # Addon Execution Initialisation
        # When the addon is run, stdout must be redirected
        # Create a buffer for redirecting stdout
        self.stdout_buffer = Queue()
        # Create a writestream - this is a wrapper for the buffer, that behaves
        # like stdout (ie it has a write function)
        self.writestream = WriteStream(self.stdout_buffer)
        # Create the object that writes to the textedit
        self.text_receiver = TextReceiver(self.stdout_buffer)
        self.text_receiver.sig_text_received.connect(self.output.append)

        ### Should just create a Thread once???
        # Create a thread for the receiver
        self.receiver_thread = QThread()
        # Run the receiver
        self.receiver_thread.started.connect(self.text_receiver.run)
        # Move the receiver to the thread
        self.text_receiver.moveToThread(self.receiver_thread)


    def init_ui(self):
        '''
        Initialise the UI components
        '''
        self.layout = QVBoxLayout()
        self.setLayout(self.layout)

        writer_btn = QPushButton("Open Addon Writer")
        writer_btn.clicked.connect(self.open_writer)
        self.layout.addWidget(writer_btn)

        search_hbox = QHBoxLayout()
        search_label = QLabel("Search:")
        search_hbox.addWidget(search_label)
        self.search = QLineEdit(self)
        search_hbox.addWidget(self.search)
        self.layout.addLayout(search_hbox)

        self.tree = QTreeWidget(self)
        self.tree.setHeaderLabels(["Name", "Author", "Description"])
        self.import_export = QTreeWidgetItem(self.tree)
        self.import_export.setText(0, "Import/Export")
        self.analysis = QTreeWidgetItem(self.tree)
        self.analysis.setText(0, "Analysis")
        self.plotting = QTreeWidgetItem(self.tree)
        self.plotting.setText(0, "Plotting")
        self.tree.itemDoubleClicked.connect(self.run_selected)
        self.layout.addWidget(self.tree)

        self.load_btn = QPushButton(self)
        self.load_btn.setText("Load Addon")
        self.load_btn.clicked.connect(self.load_new)
        self.layout.addWidget(self.load_btn)

        self.run_btn = QPushButton(self)
        self.run_btn.setText("Run selected")
        self.run_btn.clicked.connect(self.run_selected)
        self.layout.addWidget(self.run_btn)

        output_label = QLabel("Output:")
        self.layout.addWidget(output_label)
        self.output = QTextEdit(self)
        self.output.setReadOnly(True)
        self.layout.addWidget(self.output)


    def discover_addons(self, path):
        """Find any addons contained in path and load them"""
        print("Discovering addons...")
        if os.path.exists(path):
            path_to_search = path
        elif os.path.exists('./addons/'):
            print("Path does not exist. Searching ./addons instead.")
            path_to_search = './addons/'
        else:
            print("Path does not exist. Searching current directory instead.")
            path_to_search = './'

        addon_list = []

        for file in os.listdir(path_to_search):
            if file.endswith(".py"):
                with open(path_to_search + file, 'r') as f:
                    line1 = f.readline()
                if line1 == "#cued_datalogger_addon\n":
                    addon_list.append(path_to_search+file)

        if addon_list:
            print("\t Found:")
            for addon in addon_list:
                print("\t\t " + addon)
                self.add_addon(addon)
        else:
            print("None found.")

    def load_new(self):
        # Get a list of URLs from a QFileDialog
        url_list = QFileDialog.getOpenFileNames(self, "Load new addon", "addons",
                                               "CUED DataLogger Addons (*.py)")[0]

        # For each url, add it to the tree
        for url in url_list:
            self.add_addon(url)

    def add_addon(self, addon_url):
        # Read the file
        try:
            with open(addon_url) as a:
                # Execute the file
                # WARNING: THIS IS RATHER DANGEROUS - there could be anything
                # in there!
                exec(a.read(), self.addon_local_vars, self.addon_global_vars)
        except:
            print('Error detected in code!')
            t,v,tb = sys.exc_info()
            print(t)
            print(v)
            print(traceback.format_tb(tb))
            return

        # Extract the metadata
        metadata = self.addon_global_vars["addon_metadata"]

        # Add the addon to the tree
        if metadata["category"] == "Import/Export":
            parent = self.import_export
        elif metadata["category"] == "Analysis":
            parent = self.analysis
        elif metadata["category"] == "Plotting":
            parent = self.plotting
        else:
            parent = self.tree

        addon_item = QTreeWidgetItem(parent, [metadata["name"],
                                              metadata["author"],
                                              metadata["description"]])

        # Save the addon function
        self.addon_functions[metadata["name"]] = self.addon_global_vars["run"]

    def run_selected(self):
        # Get the addon metadata from the tree
        name = self.tree.currentItem().data(0, Qt.DisplayRole)
        author = self.tree.currentItem().data(1, Qt.DisplayRole)
        description = self.tree.currentItem().data(2, Qt.DisplayRole)
        # # Run the addon function, but redirect the stdout
        # Redirect stdout to the writestream
        stdout_old = sys.stdout
        sys.stdout = self.writestream

        # Start the receiver
        self.start_receiver_thread()

        # Print some info about the addon
        print("###\n {} by {}\n {}\n###".format(name, author, description))
        # Execute the addon
        try:
            self.addon_functions[name](self.parent)
        except:
            t,v,tb = sys.exc_info()
            print(t)
            print(v)
            print(traceback.format_tb(tb))
            print('Error in Code!')
        # Tidy up
        # TODO maybe need to have a more robust way of ensuring that everything
        # closes properly eg. if application quits before these lines reached
        self.receiver_thread.quit()
        sys.stdout = stdout_old

    def start_receiver_thread(self):
        self.receiver_thread.start()

    def open_writer(self):
        if not self.addon_writer:
            self.addon_writer = AddonWriter()
            self.addon_writer.done.connect(self.done_writer)
        else:
            self.addon_writer.show()

    def done_writer(self):
        self.addon_writer.done.disconnect()
        self.addon_writer = None

class WriteStream(object):
    """A simple object that writes to a queue - replace stdout with this"""
    def __init__(self, queue):
        self.queue = queue

    def write(self, text):
        self.queue.put(text)


class TextReceiver(QObject):
    """Sits blocking until data is written to stdout_buffer, which it then
    emits as a signal. To be run in a QThread."""
    sig_text_received = pyqtSignal(str)

    def __init__(self, stdout_buffer):
        super().__init__()
        self.stdout_buffer = stdout_buffer

    def run(self):
        while True:
            # Get text from stdout (block until there's something to get)
            # If we got something, send the text received signal
            # Blocking actually crashes the program :(
            stdout_text = self.stdout_buffer.get(block = False)
            self.sig_text_received.emit(stdout_text)

                #break

class AddonWriter(QWidget):
    '''
    Opens a window to write an addon, following the template. Still in alpha stage.
    The window is also able to load an addon for editting.

    Attributes
    ----------
    done: pyqtsignal
        Emits when the window is closed
    '''
    done = pyqtSignal()
    def __init__(self):
        super().__init__()
        self.setGeometry(500,300,600,500)
        self.setWindowTitle('AddonWriter')
        self.init_UI()
        self.show()

    def init_UI(self):
        '''
        Construct the window
        '''
        main_layout = QVBoxLayout(self)

        btn_layout = QHBoxLayout()
        save_btn = QPushButton('Save',self)
        save_btn.clicked.connect(self.create_addon)
        btn_layout.addWidget(save_btn)
        load_btn = QPushButton('Load',self)
        load_btn.clicked.connect(self.read_addon)
        btn_layout.addWidget(load_btn)
        run_btn = QPushButton('Run',self)
        run_btn.setDisabled(True)
        btn_layout.addWidget(run_btn)
        clear_btn = QPushButton('Clear',self)
        clear_btn.setDisabled(True)
        btn_layout.addWidget(clear_btn)

        main_layout.addLayout(btn_layout)

        self.tabs = QTabWidget(self)
        main_layout.addWidget(self.tabs)

        metadata_widget = QWidget(self)
        m_widget_layout = QVBoxLayout(metadata_widget)
        title_label = QLabel('Code Metadata')
        m_widget_layout.addWidget(title_label)
        metadata_layout = QFormLayout()

        self.meta_configs = []
        for mdata in ("File Name","Name","Author","Category"):
            if mdata == "Category":
                cbox = QComboBox(metadata_widget)
                cbox.addItems(["Import/Export","Analysis","Plotting"])
            else:
                cbox = QLineEdit(metadata_widget)
            metadata_layout.addRow(QLabel(mdata,metadata_widget),cbox)
            self.meta_configs.append(cbox)
        m_widget_layout.addLayout(metadata_layout)

        desc_label = QLabel('Description')
        m_widget_layout.addWidget(desc_label)

        desc_box = QTextEdit(metadata_widget)
        m_widget_layout.addWidget(desc_box)
        self.meta_configs.append(desc_box)

        code_widget = QWidget(self)
        c_widget_layout = QVBoxLayout(code_widget)

        font_metrics = QFontMetrics(QFont())
        w = font_metrics.width(' ')

        main_code_label = QLabel('Run Code',code_widget)
        self.main_code_text = QTextEdit(code_widget)
        self.main_code_text.setTabStopWidth(w*4)

        c_widget_layout.addWidget(main_code_label)
        c_widget_layout.addWidget(self.main_code_text)

        self.tabs.addTab(metadata_widget,'Metadata')
        self.tabs.addTab(code_widget,'Code')

    def closeEvent(self,event):
        '''
        Reimplemented from QWidget to emit the done signal
        '''
        self.done.emit()
        event.accept()
        self.deleteLater()

    def create_addon(self):
        '''
        Create the addon in the given template
        '''
        file_name = self.meta_configs[0].text().replace(' ','')
        if not file_name:
            print('Please give a file name')
            return
        metadata = [self.meta_configs[1].text(),
                    self.meta_configs[2].text(),
                    self.meta_configs[3].currentText(),
                    self.meta_configs[4].toPlainText()]
        metadata[3] = metadata[3].replace("\n"," ")
        metadata = ["\"" + md + "\""  for md in metadata]

        with open("./addons/%s.py" % file_name,'w',encoding='ASCII') as file:
            file.write('#cued_datalogger_addon\n')
            file.write("""addon_metadata = {"name": %s,\n"author": %s,\n"category": %s,\n"description": %s}\n\n""" %
                       tuple(metadata))
            main_code = self.main_code_text.toPlainText()
            if not main_code:
                main_code = 'pass'
            full_code = "def run(parent_window):\n" + main_code
            full_code = full_code.replace('\n','\n\t')
            file.write(full_code)

    def read_addon(self):
        '''
        Read the addon
        '''
        url_list = QFileDialog.getOpenFileNames(self, "Load addon", "addons",
                                               "DataLogger Addons (*.py)")[0][0]
        
        if url_list:
            f = QFileInfo(url_list)
            self.meta_configs[0].setText(f.fileName().strip('.py'))
            with open(url_list) as file:
                data = file.readlines()
            func_re = re.compile(r'def run(\S*):')
            func_search = [True if func_re.search(s) else False for s in data ].index(True)
            self.main_code_text.clear()
            indentRE = re.compile(r'(\s){4}|\t')
            for line in data[func_search+1:]:
                self.main_code_text.append(indentRE.sub('',line.strip('\n'),1))

            try:
                addon_local_vars = {}
                addon_global_vars = {}
                with open(url_list) as file:
                    exec(file.read(), addon_local_vars, addon_global_vars)
            except:
                print('Error detected in code!')
                t,v,tb = sys.exc_info()
                print(t)
                print(v)
                print(traceback.format_tb(tb))
                return
            
            metadata = addon_global_vars["addon_metadata"]
            # Extract the metadata
            self.meta_configs[1].setText(metadata["name"])
            self.meta_configs[2].setText(metadata["author"])
            self.meta_configs[3].setCurrentText(metadata["category"])
            self.meta_configs[4].setText(metadata["description"])

        else:
            print('No data')

'''
if __name__ == '__main__':
    app = 0
    app = QApplication(sys.argv)

    w = AnalysisWindow()
    w.addon_widget.discover_addons("../addons/")

    w.show()

    sys.exit(app.exec_())
'''


# -*- coding: utf-8 -*-
"""
Created on Thu Aug 24 17:35:00 2017

@author: eyt21

This module contains the live graph classes to the acquisition window.

Attributes
----------
CHANLVL_FACTOR: float
    Not used
TRACE_DECAY: float
    The decay factor of the peak plots
TRACE_DURATION: float
    Duration before the peak plots decay
"""
from PyQt5.QtGui import QColor
from PyQt5.QtCore import Qt, pyqtSignal

from cued_datalogger.api.pyqtgraph_extensions import CustomPlotWidget

import pyqtgraph as pg
import numpy as np
import math

CHANLVL_FACTOR = 0.1
TRACE_DECAY = 0.005
TRACE_DURATION = 2.0

class LiveGraph(pg.PlotWidget):
    """
    A base PlotWidget reimplemented to store extra plot information, such as
    offsets, colours, and visibility.

    Attributes
    ----------
    plotColourChanged: pyqtsignal
        Emits when colour of a plot change
        Sends out QColor
    plotlines: list of PlotDataItem
        Contains the individual PlotDataItem
    plot_xoffset: list of float
        Contains the X offset of each plot
    plot_yoffset: list of float
        Contains the Y offset of each plot
    plot_colours:list of QColor
        Contains the current colour of each plot
    plot_visible: list of bool
        Contains the visibility of each plot
    """
    #plotLineClicked = pyqtSignal()
    plotColourChanged = pyqtSignal(object)
    def __init__(self,*args,**kwargs):
        """
        Reimplemented PlotWidget
        Set the background black and axes white
        All parameters are passed into PlotWidget
        """
        super().__init__(*args,background = 'k',**kwargs)
        self.plotlines = []
        self.plot_xoffset = []
        self.plot_yoffset = []
        self.plot_colours = []
        self.plot_visible = []
        self.plotItem = self.getPlotItem()
        self.gen_default_colour()

        self.plotItem.getAxis('bottom').setPen('w')
        self.plotItem.getAxis('left').setPen('w')

    def plot(self, *arg, **kwargs):
        """
        Plot the data and set it to be clickable

        Returns
        ----------
        PlotDataItem
            The plot line, effectively
        """
        line = self.plotItem.plot(*arg, **kwargs)
        line.curve.setClickable(True,width = 4)
        self.plotlines.append(line)
        return line

    def check_line(self,line):
        """
        Check whether a plot line exists

        Returns
        ----------
        int
            Index of the plot line, if it exists
            None otherwise
        """
        if line in self.plotlines:
            return self.plotlines.index(line)
        else:
            return None

    def toggle_plotline(self,num,visible):
        """
        Set the visibility of the specific line

        Parameters
        ----------
        num: int
            index of the line to be set
        visible: bool
            Visibility of the line to be set
        """
        self.plot_visible[num] = visible
        if visible:
            self.plotlines[num].setPen(self.plot_colours[num])
        else:
            self.plotlines[num].setPen(None)

    def set_plot_colour(self,num,col):
        """
        Set the colour of the specific line

        Parameters
        ----------
        num: int
            index of the line to be set
        col: QColor
            Colour of the line to be set
        """
        self.plot_colours[num] = col
        if self.plot_visible[num]:
            self.plotlines[num].setPen(col)

    def set_offset(self,num,x_off = None, y_off =None):
        """
        Set the offsets of the specific line

        Parameters
        ----------
        num : int
            Index of the line to be set
        x_off : float
            X offset of the line to be set, if given a value
        x_off : float
            Y offset of the line to be set, if given a value
        """
        if not x_off is None:
            self.plot_xoffset[num] = x_off
        if not y_off is None:
            self.plot_yoffset[num] = y_off

    def update_line(self, num, x=None, y=None, *arg, **kwargs):
        """
        Update the existing lines with new data, with the offsets

        Parameters
        ----------
        num : int
            index of the line to be set
        x : float
            X data of the line to be set, if given a value
        y : float
            Y data of the line to be set, if given a value
        The rest to pass to PlotDataItem.setData
        """
        self.plotlines[num].setData(*arg,x = x+self.plot_xoffset[num],
                      y = y+ self.plot_yoffset[num],**kwargs)

    def reset_plotlines(self):
        """
        Clear all of the lines
        """
        for _ in range(len(self.plotlines)):
            line = self.plotlines.pop()
            line.clear()
            del line

    def reset_offsets(self):
        """
        Reset the offsets of the plots
        """
        n = len(self.plotlines)
        self.plot_xoffset = np.zeros(shape = (n,), dtype = np.float)
        self.plot_yoffset = np.arange(n, dtype = np.float)

    def reset_plot_visible(self):
        """
        Reset the visibilities of the plots
        """
        self.plot_visible = [True]*len(self.plotlines)

    def reset_colour(self):
        """
        Clear the colours of the plots
        """
        self.plot_colours = [None] * len(self.plotlines)
        self.gen_default_colour()

    def reset_default_colour(self,num):
        """
        Set the default colour of the specified plot

        Parameters
        ----------
        num: int
            Index of the line to be set
        """
        col = self.def_colours[num]
        self.set_plot_colour(num,col)
        self.plotColourChanged.emit(col)
        #return col

    def gen_default_colour(self):
        """
        Generate the default colours of the plots
        """
        val = [0.0,0.5,1.0]
        colour = np.array([[255,0,0,255],[0,255,0,255],[0,0,255,255]], dtype = np.ubyte)
        self.plot_colourmap =  pg.ColorMap(val,colour)
        c_list = self.plot_colourmap.getLookupTable(nPts = len(self.plotlines))

        self.def_colours = []
        for i in range(len(self.plotlines)):
            r,g,b = c_list[i]
            self.set_plot_colour(i,QColor(r,g,b))
            self.def_colours.append(QColor(r,g,b))

class TimeLiveGraph(LiveGraph):
    """
    Reimplemented LiveGraph. Displays the time domain plot

    Attributes
    ----------
    sig_hold: list of bool
        Contains whether the signal is being held
    """
    def __init__(self, *args,**kwargs):
        """
        Reimplemented from LiveGraph.
        """
        super().__init__(*args,**kwargs)
        self.sig_hold = []
        self.plotItem.setTitle(title="Time Plot", color = 'FFFFFF')
        self.plotItem.setLabel('bottom','Time(s)')

    def set_sig_hold(self, num, state):
        """
        Set the hold status of the specific line

        Parameters
        ----------
        num: int
            Index of the line to be set
        state: bool
            Hold status of the line to be set
        """
        self.sig_hold[num] = state

    def reset_sig_hold(self):
        self.sig_hold = [Qt.Unchecked] * len(self.plotlines)

class FreqLiveGraph(LiveGraph):
    """
    Reimplemented LiveGraph. Displays the frequency domain plot
    """
    def __init__(self,*args,**kwargs):
        """
        Reimplemented from LiveGraph.
        """
        super().__init__(*args,**kwargs)
        self.plotItem.setTitle(title="FFT Plot", color = 'FFFFFF')
        self.plotItem.setLabel('bottom','Freq(Hz)')
        self.plotItem.disableAutoRange(axis=None)


class LevelsLiveGraph(LiveGraph):
    """
    Reimplemented LiveGraph. Displays the channel levels

    Attributes
    ----------
    thresholdChanged: pyqtSignal
        Emits when the threshold line is moved
        Sends out the value of the threshold
    peak_plots: list of plotDataItem
        The lines which indicate the channels' peaks
    peak_trace: list of float
        The values of the channels' peaks
    trace_counter: list of int
        Counter for the peak plots before they decay
    chanlvl_pts: list of plotDataItem
        Rms plots
    chanlvl_bars: list of bool
        Instantaneous channels' peaks plots
    threshold_line:
        The line indicating the trigger threshold
    level_colourmap:
        The colour for the peak levels
    """
    thresholdChanged = pyqtSignal(str)
    def __init__(self,rec,*args,**kwargs):
        """
        Reimplemented from LiveGraph.

        Parameters
        ----------
        rec: Recorder
            The reference of the Recorder
        The rest are passed into LiveGraph
        """
        self.peak_plots = []
        self.peak_trace = []
        self.trace_counter = []
        self.trace_countlimit = 30
        self.level_colourmap = None

        super().__init__(*args,**kwargs)
        self.plotItem.setTitle(title="Channel Levels", color = 'FFFFFF')
        self.plotItem.setLabel('bottom','Amplitude')
        self.plotItem.hideAxis('left')

        self.chanlvl_pts = self.plotItem.plot()
        self.chanlvl_bars = pg.ErrorBarItem(x=np.arange(rec.channels),
                                            y =np.arange(rec.channels)*0.1,
                                            beam = CHANLVL_FACTOR/2,
                                            pen = pg.mkPen(width = 3))

        self.plotItem.addItem(self.chanlvl_bars)

        baseline = pg.InfiniteLine(pos = 0.0, movable = False)
        self.plotItem.addItem(baseline)

        self.threshold_line = pg.InfiniteLine(pos = 0.0, movable = True,bounds = [0,1])
        self.threshold_line.sigPositionChanged.connect(self.change_threshold)
        self.plotItem.addItem(self.threshold_line)

        self.reset_channel_peaks(rec)

        val = [0.0,0.5,0.8]
        colour = np.array([[0,255,0,255],[0,255,0,255],[255,0,0,255]], dtype = np.ubyte)
        self.level_colourmap = pg.ColorMap(val,colour)

    def set_plot_colour(self,num,col):
        """
        Parameters
        ----------
        num: int
            index of the point to be set
        col: QColor
            Colour of the point to be set
        """
        self.plot_colours[num] = col
        self.chanlvl_pts.scatter.setBrush(col)

    def set_peaks(self,num,maximum):
        """
        Set the value of the peak plots
        Parameters
        ----------
        num: int
            index of the peak to be set
        maximum: float
            Instantaneous maximum value of the peak
        """
        if self.trace_counter[num]>self.trace_countlimit:
            self.peak_trace[num] = max(self.peak_trace[num]*math.exp(-self.peak_decays[num]),0)
            self.peak_decays[num] += TRACE_DECAY
        self.trace_counter[num] += 1

        if self.peak_trace[num]<maximum:
            self.peak_trace[num] = maximum
            self.peak_decays[num] = 0
            self.trace_counter[num] = 0

        self.peak_plots[num].setData(x = [self.peak_trace[num],self.peak_trace[num]],
                       y = [(num-0.3), (num+0.3)])

        self.peak_plots[num].setPen(self.level_colourmap.map(self.peak_trace[num]))

    def set_channel_levels(self,value,maximum):
        """
        Set the value of the levels plots
        Parameters
        ----------
        value: float
            rms values
        maximum: float
            Instantaneous maximum value of the plot
        """
        self.chanlvl_bars.setData(x = value,y = np.arange(len(self.peak_plots)), right = maximum-value,left = value)
        self.chanlvl_pts.setData(x = value,y = np.arange(len(self.peak_plots)))

    def change_threshold(self,arg):
        """
        Set the trigger threshold
        If arg is str, set the threshold_line to match the value
        otherwise, emit the value of the threshold_line

        Parameters
        ----------
        arg: str or InfiniteLine
        """
        if type(arg) == str:
            self.threshold_line.setValue(float(arg))
        else:
            self.thresholdChanged.emit('%.2f' % arg.value())

    def reset_colour(self):
        """
        Reimplemented from LiveGraph.
        """
        self.plot_colours = [None] * len(self.peak_plots)
        self.gen_default_colour()

    def reset_default_colour(self,chan):
        """
        Reimplemented from LiveGraph.
        """
        col = self.def_colours[chan]
        self.set_plot_colour(chan,col)
        return col

    def gen_default_colour(self):
        """
        Reimplemented from LiveGraph.
        """
        val = [0.0,0.5,1.0]
        colour = np.array([[255,0,0,255],[0,255,0,255],[0,0,255,255]], dtype = np.ubyte)
        plot_colourmap =  pg.ColorMap(val,colour)
        c_list = plot_colourmap.getLookupTable(nPts = len(self.peak_plots))

        self.def_colours = []
        for i in range(len(self.peak_plots)):
            r,g,b = c_list[i]
            #self.plotlines.set_plot_colour(i,QColor(r,g,b),True)
            self.plot_colours[i] = QColor(r,g,b)
            self.def_colours.append(QColor(r,g,b))

    def reset_channel_levels(self):
        """
        Reset the channel levels plot
        """
        self.chanlvl_pts.clear()
        self.chanlvl_pts = self.plotItem.plot(pen = None,symbol='o',
                                                  symbolBrush = self.plot_colours,
                                                  symbolPen = None)

    def reset_channel_peaks(self,rec):
        """
        Reset the channel peaks plot
        """
        for _ in range(len(self.peak_plots)):
            line = self.peak_plots.pop()
            line.clear()
            del line

        self.peak_trace = np.zeros(rec.channels)
        self.peak_decays = np.zeros(rec.channels)
        self.trace_counter = np.zeros(rec.channels)
        self.trace_countlimit = TRACE_DURATION *rec.rate//rec.chunk_size

        self.threshold_line.setBounds((0,rec.max_value))

        for i in range(rec.channels):
            self.peak_plots.append(self.plotItem.plot(x = [self.peak_trace[i],self.peak_trace[i]],
                                                          y = [(i-0.3), (i+0.3)]))


        self.plotItem.setRange(xRange = (0,rec.max_value+0.1),yRange = (-0.5, (rec.channels+5-0.5)))
        self.plotItem.setLimits(xMin = -0.1,xMax = rec.max_value+0.1,yMin = -0.5,yMax = (rec.channels+5-0.5))



# -*- coding: utf-8 -*-
"""
Created on Tue Aug 22 11:19:29 2017
@author: eyt21

This module contains the widget classes to the acquisition window.
However, they are not limited to that window, and can be reused for other
window, like the analysis window.

Attributes
----------
NI_DRIVERS: bool
    Indicates whether NIDAQmx drivers and pyDAQmx module are installed
    when attempting to import NIRecorder module
    The module is needed to check on the available National Instrument devices
MAX_SAMPLE: int
    Arbritrary maximum number of samples that can be recorded.

"""
from PyQt5.QtWidgets import (QWidget, QVBoxLayout, QHBoxLayout, QPushButton,
                             QStatusBar, QLabel, QLineEdit, QFormLayout,
                             QGroupBox, QRadioButton, QComboBox, QScrollArea,
                             QGridLayout, QCheckBox, QButtonGroup,
                             QStackedWidget)
from PyQt5.QtGui import QValidator,QIntValidator,QDoubleValidator,QPainter
from PyQt5.QtCore import Qt, pyqtSignal
from PyQt5.Qt import QStyleOption,QStyle

import re
import pyqtgraph as pg
import functools as fct

import cued_datalogger.acquisition.myRecorder as mR
try:
    import cued_datalogger.acquisition.NIRecorder as NIR
    NI_drivers = True
except NotImplementedError:
    print("Seems like you don't have National Instruments drivers")
    NI_drivers = False
except ImportError:
    print("ImportError: Seems like you don't have pyDAQmx modules")
    NI_drivers = False

MAX_SAMPLE = 1e9

#==========================WIDGET CLASSES================================
#---------------------------BASE WIDGET------------------------------------
class BaseWidget(QWidget):
    """
     A base widget reimplemented to allow custom styling.
     Pretty much identical to a normal QWidget
    """
    def __init__(self, *arg, **kwarg):
        """
        Reimplemented from QWidget
        """
        super().__init__(*arg, **kwarg)
        self.initUI()
        self.setObjectName('subWidget')

    def paintEvent(self, evt):
        """
        Reimplemented from QWidget.paintEvent()
        """
        super().paintEvent(evt)
        opt = QStyleOption()
        opt.initFrom(self)
        p = QPainter(self)
        s = self.style()
        s.drawPrimitive(QStyle.PE_Widget, opt, p, self)

    def initUI(self):
        """
        Construct the UI, to be reimplemented.
        """
        pass

#-------------------------CHANNEL TOGGLE WIDGET-------------------------------
class ChanToggleUI(BaseWidget):
    """
    A Channel Toggling widget.
    Contains:
        * Checkboxes to toggle channel,
        * Buttons to select all, deselect all, and invert selection.
        * LineEdits to toggle by expression or tags

    Attributes
    ----------
    toggleChanged: pyqtsignal
        Emits when a channel toggle changes,
        Sends out the channel num(int) and the state(bool)
    channels_box: QWidget
        The widget containing the checkboxes
    checkbox_layout: QGridLayout
        Defines the layout of the checkboxes
    chan_btn_group: QButtonGroup
        Widget to handle the checkboxes presses
    sel_all_btn:QPushButton
        'Select All' button
    desel_all_btn: QPushButton
        'Deselect All' button
    inv_sel_btn: QPushButton
        'Invert Selection' button
    chan_text: ChanLineText
        For toggling by expression
    chan_text2: QLineEdit
        For toggling by tags
    chan_text3: QLineEdit
        For displaying the channels toggled (may be changed to QLabel instead)
    search_status: QStatusBar
        For displaying whether the toggling is successful
    """
    sigToggleChanged = pyqtSignal(int,bool)

    def initUI(self):
        """
        Reimplemented from BaseWidget.
        """
        # Set up the channel tickboxes widget
        chans_toggle_layout = QVBoxLayout(self)

        # Make the button tickboxes scrollable
        scroll = QScrollArea()

        self.channels_box = QWidget(scroll)
        self.checkbox_layout = QGridLayout(self.channels_box)

        # Set up the QbuttonGroup to manage the Signals
        self.chan_btn_group = QButtonGroup(self.channels_box)
        self.chan_btn_group.setExclusive(False)

        scroll.setWidget(self.channels_box)
        scroll.setWidgetResizable(True)

        chans_toggle_layout.addWidget(scroll)

        # Set up the selection toggle buttons
        self.sel_all_btn = QPushButton('Select All', self)
        self.desel_all_btn = QPushButton('Deselect All',self)
        self.inv_sel_btn = QPushButton('Invert Selection',self)
        for btn in [self.sel_all_btn,self.desel_all_btn,self.inv_sel_btn]:
            btn.resize(btn.sizeHint())
            chans_toggle_layout.addWidget(btn)

        self.sel_all_btn.clicked.connect(lambda: self.toggle_all_checkboxes(Qt.Checked))
        self.desel_all_btn.clicked.connect(lambda: self.toggle_all_checkboxes(Qt.Unchecked))
        self.inv_sel_btn.clicked.connect(self.invert_checkboxes)

        self.chan_text = ChanLineText(self)
        self.chan_text2 = QLineEdit(self)
        self.chan_text3 = QLineEdit(self)
        self.search_status = QStatusBar(self)
        self.search_status.setSizeGripEnabled(False)
        code_warning = QLabel('**Toggling by expression or tags**')
        code_warning.setWordWrap(True)
        chans_toggle_layout.addWidget(code_warning)
        chans_toggle_layout.addWidget(QLabel('Expression:'))
        chans_toggle_layout.addWidget(self.chan_text)

        chans_toggle_layout.addWidget(QLabel('Hashtag Toggle:'))
        chans_toggle_layout.addWidget(self.chan_text2)
        chans_toggle_layout.addWidget(QLabel('Channel(s) Toggled:'))
        chans_toggle_layout.addWidget(self.chan_text3)
        chans_toggle_layout.addWidget(self.search_status)

        self.search_status.showMessage('Awaiting...')

        self.chan_btn_group.buttonClicked.connect(self.toggle_channel_plot)
        self.chan_text.returnPressed.connect(self.chan_line_toggle)

    def invert_checkboxes(self):
        """
        Callback to invert selection
        """
        for btn in self.channels_box.findChildren(QCheckBox):
            btn.click()

    def toggle_all_checkboxes(self, state):
        """
        Callback to select all or deselect all

        Parameters
        ----------
        state: int
            State of the checkboxes to be in (either Qt.Unchecked or Qt.Checked)
        """
        for btn in self.channels_box.findChildren(QCheckBox):
            if not btn.checkState() == state:
                btn.click()

    def adjust_channel_checkboxes(self, new_n_btns):
        """
        Add or delete checkboxes based on new number of buttons needed

        Parameters
        ----------
        new_n_btns: int
            New number of buttons required
        """
        for btn in self.chan_btn_group.buttons():
            btn.setCheckState(Qt.Checked)

        # Calculate the number of extra buttons required to add/delete
        old_n_buttons = self.checkbox_layout.count()
        extra_btns = abs(new_n_btns - old_n_buttons)
        # Add/Delete the extra buttons if there are extras
        if extra_btns:
            if new_n_btns > old_n_buttons:
                columns_limit = 4
                current_y = (old_n_buttons-1)//columns_limit
                current_x = (old_n_buttons-1)%columns_limit
                for n in range(old_n_buttons,new_n_btns):
                    current_x +=1
                    if current_x%columns_limit == 0:
                        current_y +=1
                    current_x = current_x%columns_limit

                    chan_btn = QCheckBox('Channel %i' % n,self.channels_box)
                    chan_btn.setCheckState(Qt.Checked)
                    self.checkbox_layout.addWidget(chan_btn,current_y,current_x)
                    self.chan_btn_group.addButton(chan_btn,n)
            else:
                for n in range(old_n_buttons-1,self.rec.new_n_btns-1,-1):
                    chan_btn = self.chan_btn_group.button(n)
                    self.checkbox_layout.removeWidget(chan_btn)
                    self.chan_btn_group.removeButton(chan_btn)
                    chan_btn.deleteLater()

    def chan_line_toggle(self, chan_list):
        """
        Callback to intepret the input expressions and toggle the channels
        accordingly

        Parameters
        ----------
        chan_list: List of str
            Input expressions
        """
        # intepret the string into either int or range, then append/extend the all_selected_chan list
        all_selected_chan = []
        for str_in in chan_list:
            r_in = str_in.split(':')
            if len(r_in) == 1:
                if r_in[0]:
                    all_selected_chan.append(int(r_in[0]))
            elif len(r_in) >1:
                if not r_in[0]:
                    r_in[0] = 0
                if not r_in[1]:
                    r_in[1] = self.rec.channels

                if len(r_in) == 2:
                    all_selected_chan.extend(range(int(r_in[0]),int(r_in[1])+1))
                else:
                    if not r_in[2]:
                        r_in[2] = 1
                    all_selected_chan.extend(range(int(r_in[0]),int(r_in[1])+1,int(r_in[2])))

        # Toggle checkboxes based on the integers in the list
        print(all_selected_chan)
        if all_selected_chan:
            n_btns = self.checkbox_layout.count()
            self.toggle_all_checkboxes(Qt.Unchecked)
            for chan in set(all_selected_chan):
                if chan < n_btns:
                    self.chan_btn_group.button(chan).click()

    def toggle_channel_plot(self, btn):
        """
        Callback when a checkbox is clicked.
        Emits sigToggleChanged.

        Parameters
        ----------
            btn: QCheckBox
                button that is clicked on
        """
        chan_num = self.chan_btn_group.id(btn)
        if btn.isChecked():
            self.sigToggleChanged.emit(chan_num,True)
        else:
            self.sigToggleChanged.emit(chan_num,False)

#-----------------------CHANNEL CONFIGURATION WIDGET-------------------------
class ChanConfigUI(BaseWidget):
    """
    A Channel Plots Configuration widget.
    Contains:
        * ComboBox to switch channel plot info,
        * Spinboxes to set the offsets
        * Buttons to change the colour of a plot
        * Checkbox to hold a signal
        * Button to open a window to edit metadata

    Attributes
    ----------
    timeOffsetChanged: pyqtsignal
        Emits when a time domain offset is changed,
        Sends out the channel num(int) and the x and y offsets(float,float)
    freqOffsetChanged: pyqtsignal
        Emits when a frequency domain offset is changed,
        Sends out the channel num(int) and the x and y offsets(float,float)
    sigHoldChanged: pyqtsignal
        Emits when a state of holding the plot is changed,
        Sends out the channel num(int) and the state(bool)
    colourReset: pyqtsignal
        Emits when a plot colour is reset,
        Sends out the channel num(int)
    colourChanged: pyqtsignal
        Emits when a plot colour is changed,
        Sends out the channel num(int) and the color(QColor)
    chans_num_box: QComboBox
        The widget to select the channel plot
    hold_tickbox: QCheckBox
        Toggles whether to hold the signal or not
    colbox:QPushButton
        Set the colour of the plot
    defcol_btn: QPushButton
        Reset the colour of the plot to the default colour
    meta_btn: QPushButton
        Opens the metadata editing window
    time_offset_config: List of SpinBox
        Sets the X and Y offsets of time domain plot
    fft_offset_config: List of SpinBox
        Sets the X and Y offsets of frequency domain plot
    """
    sigTimeOffsetChanged = pyqtSignal(int,float,float)
    sigFreqOffsetChanged = pyqtSignal(int,float,float)
    sigHoldChanged = pyqtSignal(int,int)
    sigColourReset = pyqtSignal(int)
    sigColourChanged = pyqtSignal(int,object)

    def initUI(self):
        """
        Reimplemented from BaseWidget.
        """
        chans_prop_layout = QVBoxLayout(self)
        chans_prop_layout.setContentsMargins(5,5,5,5)
        #chans_prop_layout.setSpacing(10)
        chan_num_col_layout = QGridLayout()

        self.chans_num_box = QComboBox(self)
        self.hold_tickbox = QCheckBox('Hold',self)
        chan_num_col_layout.addWidget(QLabel('Channel',self),0,0)
        chan_num_col_layout.addWidget(self.chans_num_box,0,1)
        chan_num_col_layout.addWidget(self.hold_tickbox,0,2)

        self.colbox = pg.ColorButton(self,(0,255,0))
        self.defcol_btn = QPushButton('Reset Colour',self)
        self.meta_btn = QPushButton('Channel Info',self)
        chan_num_col_layout.addWidget(QLabel('Colour',self),1,0)
        chan_num_col_layout.addWidget(self.colbox,1,1)
        chan_num_col_layout.addWidget(self.defcol_btn,1,2)
        chan_num_col_layout.addWidget(self.meta_btn,2,0,1,3)

        chans_prop_layout.addLayout(chan_num_col_layout)

        chan_settings_layout = QVBoxLayout()
        chan_settings_layout.setSpacing(0)

        self.time_offset_config = []
        self.fft_offset_config = []

        for set_type in ('Time','DFT'):
            settings_gbox = QGroupBox(set_type, self)
            settings_gbox.setFlat(True)
            gbox_layout = QGridLayout(settings_gbox)
            row = 0
            for c in ['XMove','YMove']:
                cbox = pg.SpinBox(parent= settings_gbox, value=0.0, bounds=[None, None],step = 0.1)
                stepbox = pg.SpinBox(parent= settings_gbox, value=0.1, bounds=[None, None],step = 0.001)
                stepbox.valueChanged.connect(fct.partial(self.set_offset_step,cbox))

                gbox_layout.addWidget(QLabel(c,self),row,0)
                gbox_layout.addWidget(cbox,row,1)
                gbox_layout.addWidget(stepbox,row,2)
                if set_type == 'Time':
                    self.time_offset_config.append(cbox)
                elif set_type == 'DFT':
                    self.fft_offset_config.append(cbox)
                row += 1
            settings_gbox.setLayout(gbox_layout)
            chan_settings_layout.addWidget(settings_gbox)

        chans_prop_layout.addLayout(chan_settings_layout)

        for cbox in self.time_offset_config:
            cbox.sigValueChanging.connect(fct.partial(self.set_plot_offset,'Time'))
        for cbox in self.fft_offset_config:
            cbox.sigValueChanging.connect(fct.partial(self.set_plot_offset,'DFT'))

        self.hold_tickbox.stateChanged.connect(self.signal_hold)
        self.colbox.sigColorChanging.connect(lambda: self.set_plot_colour())
        self.defcol_btn.clicked.connect(lambda: self.set_plot_colour(True))

    def set_offset_step(self,cbox,step_val):
        """
        Sets the single step of a spinbox

        Parameters
        ----------
        cbox: SpinBox
            SpinBox to set
        step_val: float
            The new value of the single step
        """
        cbox.setSingleStep(step_val)

    def set_plot_offset(self,dtype):
        """
        Callback to set the offset.
        Emits sigTimeOffsetChanged or sigFreqOffsetChanged depending on dtype

        Parameters
        ----------
        dtype: str
            Either 'Time' of 'DFT' to indicate the time domain or frequency domain plot respectively
        """
        chan = self.chans_num_box.currentIndex()
        if dtype == 'Time':
            self.sigTimeOffsetChanged.emit(chan,self.time_offset_config[0].value(),self.time_offset_config[1].value())
        elif dtype == 'DFT':
            self.sigFreqOffsetChanged.emit(chan,self.fft_offset_config[0].value(),self.fft_offset_config[1].value())

    def signal_hold(self,state):
        """
        Callback to hold the plot.
        Emits sigHoldChanged

        Parameters
        ----------
        dtype: str
            Either 'Time' of 'DFT' to indicate the time domain or frequency domain plot respectively
        """
        chan = self.chans_num_box.currentIndex()
        self.sigHoldChanged.emit(chan,state)

    def set_colour_btn(self,col):
        """
        Set the colour of the colour button.

        Parameters
        ----------
        col: QColor
            Colour to set
        """
        self.colbox.setColor(col)

    def set_plot_colour(self,reset = False):
        """
        Set the colour of the colour button.
        Emits either sigColourReset or sigColourChanged

        Parameters
        ----------
            reset: bool
                Whether to reset the colour or not
        """
        chan = self.chans_num_box.currentIndex()
        if reset:
            self.sigColourReset.emit(chan)
        else:
            col = self.colbox.color()
            self.sigColourChanged.emit(chan,col)

#-----------------------DEVICE CONFIGURATION WIDGET-------------------------
class DevConfigUI(BaseWidget):
    """
    A Channel Plots Configuration widget.
    Contains widgets to setup the recorder

    Attributes
    ----------
    configRecorder: pyqtsignal
        Emits the configuration of the recorder is set
    typebtngroup: QButtonGroup
        Contains the buttons to select source of audio stream
        Either SoundCard or NI
    config_button: QPushButton
        Confirm the settings and set up the new recorder
    rec: Recorder object
        Reference of the Recorder object
    configboxes: List of widgets
        Widgets for the configuration settings, in order:
            ['Source','Rate','Channels','Chunk Size','Number of Chunks']
        with type, respectively:
            [QComboBox, QLineEdit, QLineEdit, QLineEdit, QLineEdit]
    """
    configRecorder = pyqtSignal()

    def initUI(self):
        """
        Reimplemented from BaseWidget.
        """
        self.rec = None
         # Set the device settings form
        config_form = QFormLayout(self)
        config_form.setSpacing(2)

        # Set up the device type radiobuttons group
        self.typegroup = QGroupBox('Input Type', self)
        typelbox = QHBoxLayout(self.typegroup)
        pyaudio_button = QRadioButton('SoundCard',self.typegroup)
        NI_button = QRadioButton('NI',self.typegroup)
        typelbox.addWidget(pyaudio_button)
        typelbox.addWidget(NI_button)
        pyaudio_button.setChecked(True)
        # Set that to the layout of the group
        self.typegroup.setLayout(typelbox)

        # Set up QbuttonGroup to manage the buttons' Signals
        self.typebtngroup = QButtonGroup(self)
        self.typebtngroup.addButton(pyaudio_button)
        self.typebtngroup.addButton(NI_button)

        config_form.addRow(self.typegroup)

        # Add the remaining settings
        configs = ['Source','Rate','Channels','Chunk Size','Number of Chunks']
        self.configboxes = []

        for c in configs:
            if c == 'Source':
                cbox = QComboBox(self)
                config_form.addRow(QLabel(c,self),cbox)
                self.configboxes.append(cbox)

            else:
                cbox = QLineEdit(self)
                config_form.addRow(QLabel(c,self),cbox)
                self.configboxes.append(cbox)

        # Add the confirm button
        self.config_button = QPushButton('Set Config', self)
        config_form.addRow(self.config_button)

        self.config_button.clicked.connect(self.configRecorder.emit)
        self.typebtngroup.buttonReleased.connect(self.display_sources)

    def set_recorder(self,recorder):
        """
        Set the recorder for reference

        Parameters
        ----------
        recorder: Recorder object
            The reference of the Recorder object
        """
        self.rec = recorder

    def config_setup(self):
        """
        Configure the inputs of the config_boxes

        Parameters
        ----------
        recorder: Recorder object
            The reference of the Recorder object
        """
        rb = self.typegroup.findChildren(QRadioButton)
        if not self.rec == None:
            if type(self.rec) == mR.Recorder:
                rb[0].setChecked(True)
            elif type(self.rec) == NIR.Recorder:
                rb[1].setChecked(True)
                
            info = [self.rec.rate,self.rec.channels,
                self.rec.chunk_size,self.rec.num_chunk]
        for cbox,i in zip(self.configboxes[1:],info):
            cbox.setText(str(i))
            
        self.display_sources()

    def display_sources(self):
        """
        Display the available sources from the type of recorder
        Either SoundCard(myRecorder) or NI(NIRecorder)
        """
        # Check which type of recorder is selected
        rb = self.typegroup.findChildren(QRadioButton)
        if rb[0].isChecked():
            selR = mR.Recorder()
        elif rb[1].isChecked():
            selR = NIR.Recorder()
        else:
            return

        # Clear and fill the source configbox
        source_box = self.configboxes[0]
        source_box.clear()
        try:
            full_device_name = []
            # 'extras' are extra details of the devices
            names,extras =  selR.available_devices()
            for name,extra in zip(names,extras):
                if type(extra) == str:
                    full_device_name.append(name + ' - ' + extra)
                else:
                    full_device_name.append(name)

            source_box.addItems(full_device_name)
        except Exception as e:
            print(e)
            source_box.addItems(selR.available_devices()[0])
            
        if not self.rec == None:
            source_box.setCurrentText(self.rec.device_name)
        del selR

    def read_device_config(self):
        """
        Display the available sources from the type of recorder
        Either SoundCard(myRecorder) or NI(NIRecorder)

        Returns
        ----------
        recType: str
            Type of recorder
        configs: list
            The configurations ['Source','Rate','Channels','Chunk Size','Number of Chunks']
            with type, respectively:[str, int, int, int, int]
        """
        # TODO: Put in QValidators
        recType =  [rb.isChecked() for rb in self.typegroup.findChildren(QRadioButton)]
        configs = []
        for cbox in self.configboxes:
            if type(cbox) == QComboBox:
                configs.append(cbox.currentIndex())
            else:
                config_input = cbox.text().strip(' ')
                try:
                    configs.append(int(float(config_input)))
                except:
                    configs.append(None)
        print(recType,configs)
        if recType[0]:
            recType = mR
        elif recType[1]:
            recType = NIR
        return(recType, configs)

#-----------------------------STATUS WIDGET-------------------------------
class StatusUI(BaseWidget):
    """
    A Status Bar widget.
    Contains:
        * QStatusBar to display the stream status
        * Button to reset the splitters
        * Button to resume/pause the stream
        * Button to grab a snapshot of the stream

    Attributes
    ----------
    statusbar: QStatusBar
        Displays the status of the stream
    resetView: QPushButton
        Reset the splitter view
    togglebtn: Recorder object
        Resume/pause the stream
    sshotbtn: List of widgets
        Grab a snapshot of the stream
    """
    def initUI(self):
        """
        Reimplemented from BaseWidget.
        """
        stps_layout = QHBoxLayout(self)

        self.statusbar = QStatusBar(self)
        self.statusbar.showMessage('streaming')
        self.statusbar.clearMessage()
        stps_layout.addWidget(self.statusbar)

        self.resetView = QPushButton('Reset View',self)
        self.resetView.resize(self.resetView.sizeHint())
        stps_layout.addWidget(self.resetView)
        self.togglebtn = QPushButton('Pause',self)
        self.togglebtn.resize(self.togglebtn.sizeHint())
        stps_layout.addWidget(self.togglebtn)
        self.sshotbtn = QPushButton('Get Snapshot',self)
        self.sshotbtn.resize(self.sshotbtn.sizeHint())
        stps_layout.addWidget(self.sshotbtn)

    def trigger_message(self):
        """
        Display a message when the recording trigger is set off
        """
        self.statusbar.showMessage('Triggered! Recording...')

#-----------------------------RECORDING WIDGET-------------------------------
class RecUI(BaseWidget):
    """
    A Recording Configuration widget.
    Contains:
        * ComboBox to change recording mode,
        * Widgets for setting up the recording:
            * Recording samples/ duration
            * Triggering
        * Additional widgets for specific recording mode:
            * Normal: None
            * Average transfer function: Buttons to undo or clear past autospectrum and crossspectrum

    Attributes
    ----------
    startRecording: pyqtsignal
        Emits when record button is pressed
    cancelRecording: pyqtsignal
        Emits when cancel button is pressed
    undoLastTfAvg: pyqtsignal
        Emits when undo last transfer function button is pressed
    clearTfAvg: pyqtsignal
        Emits when clear past transfer functions button is pressed
    switch_rec_box: QComboBox
        Switch recording options
    rec_boxes: List of Widgets
        Configurations: ['Samples','Seconds','Pretrigger','Ref. Channel','Trig. Level']
        with types    : [QLineEdit, QLineEdit, QLineEdit, QComboBox, QLineEdit]
    spec_settings_widget:QStackedWidget
        Contains the additional settings
    input_chan_box: QComboBox
        Additional settings to put input channel for average transfer function calculation
    """
    startRecording = pyqtSignal()
    cancelRecording = pyqtSignal()

    undoLastTfAvg = pyqtSignal()
    clearTfAvg = pyqtSignal()

    def initUI(self):
        """
        Reimplemented from BaseWidget.
        """
        rec_settings_layout = QVBoxLayout(self)
        global_settings_layout = QFormLayout()
        self.spec_settings_widget= QStackedWidget(self)

        rec_title = QLabel('Recording Settings', self)
        rec_title.setStyleSheet('''
                                font: 18pt;
                                ''')
        rec_settings_layout.addWidget(rec_title)

        rec_settings_layout.addLayout(global_settings_layout)
        rec_settings_layout.addWidget(self.spec_settings_widget)

        self.switch_rec_box = QComboBox(self)
        self.switch_rec_box.addItems(['Normal','TF Avg.','TF Grid','<something>'])
        self.switch_rec_box.currentIndexChanged.connect(self.spec_settings_widget.setCurrentIndex)
        global_settings_layout.addRow(QLabel('Mode',self),self.switch_rec_box)

        # Add the recording setting UIs with the Validators
        configs = ['Samples','Seconds','Pretrigger','Ref. Channel','Trig. Level']
        default_values = [None,'1.0', '200','0','0.0']
        validators = [QIntValidator(1,MAX_SAMPLE),None,QIntValidator(-1,MAX_SAMPLE),
                      None,QDoubleValidator(0,5,2)]

        self.rec_boxes = []
        for c,v,vd in zip(configs,default_values,validators):
            if c == 'Ref. Channel':
                cbox = QComboBox(self)
            else:
                cbox = QLineEdit(self)
                cbox.setText(v)
                if vd:
                    cbox.setValidator(vd)

            global_settings_layout.addRow(QLabel(c,self),cbox)
            self.rec_boxes.append(cbox)

        # Connect the sample and time input check
        self.rec_boxes[0].editingFinished.connect(lambda: self.autoset_record_config('Samples'))
        self.rec_boxes[1].editingFinished.connect(lambda: self.autoset_record_config('Time'))
        self.rec_boxes[2].editingFinished.connect(lambda: set_input_limits(self.rec_boxes[2],-1,self.rec.chunk_size,int))
        self.rec_boxes[2].textEdited.connect(self.toggle_trigger)

        self.normal_rec = QWidget(self)
        normal_rec_layout = QVBoxLayout(self.normal_rec)
        normal_rec_layout.addWidget(QLabel('No Additional Options',self))
        self.spec_settings_widget.addWidget(self.normal_rec)

        # Widgets for average transfer function
        self.tfavg_rec = QWidget(self)
        tfavg_rec_layout = QVBoxLayout(self.tfavg_rec)
        tfavg_settings = QFormLayout(self.tfavg_rec)
        self.input_chan_box = QComboBox(self)
        tfavg_settings.addRow(QLabel('Input',self),self.input_chan_box)
        avg_layout = QHBoxLayout()
        #self.avg_input_box = QLineEdit(self)
        self.avg_count_box = QLabel('Count: 0',self)
        #avg_layout.addWidget(self.avg_input_box)
        avg_layout.addWidget(self.avg_count_box)
        tfavg_settings.addRow(QLabel('Averages',self),avg_layout)
        tfavg_rec_layout.addLayout(tfavg_settings)
        # Buttons for average transfer function
        tflog_btn_layout = QHBoxLayout()
        self.undo_log_btn = QPushButton('Undo Last',self)
        self.undo_log_btn.clicked.connect(self.undoLastTfAvg.emit)
        tflog_btn_layout.addWidget(self.undo_log_btn)
        self.clear_log_btn = QPushButton('Clear',self)
        self.clear_log_btn.clicked.connect(self.clearTfAvg.emit)
        tflog_btn_layout.addWidget(self.clear_log_btn)
        tfavg_rec_layout.addLayout(tflog_btn_layout)

        self.spec_settings_widget.addWidget(self.tfavg_rec)

        # Widgets for average transfer function grid
        self.tfgrid_rec = QWidget(self)
        tfgrid_rec_layout = QVBoxLayout(self.tfgrid_rec)
        tfgrid_rec_layout.addWidget(QLabel('Nothing is here :(',self))
        self.spec_settings_widget.addWidget(self.tfgrid_rec)

        # Placeholder
        self.something_rec = QWidget(self)
        something_rec_layout = QVBoxLayout(self.something_rec)
        something_rec_layout.addWidget(QLabel('<something is here>',self))
        self.spec_settings_widget.addWidget(self.something_rec)

        # Add the record and cancel buttons
        rec_btn_layout = QHBoxLayout(self.normal_rec)
        self.recordbtn = QPushButton('Log',self)
        self.recordbtn.resize(self.recordbtn.sizeHint())
        self.recordbtn.clicked.connect(self.startRecording.emit)
        rec_btn_layout.addWidget(self.recordbtn)
        self.cancelbtn = QPushButton('Cancel',self)
        self.cancelbtn.resize(self.cancelbtn.sizeHint())
        self.cancelbtn.setDisabled(True)
        self.cancelbtn.clicked.connect(self.cancelRecording.emit)
        rec_btn_layout.addWidget(self.cancelbtn)
        rec_settings_layout.addLayout(rec_btn_layout)

    def set_recorder(self,recorder):
        """
        Set the recorder reference
        """
        self.rec = recorder
        self.reset_configs()
        self.autoset_record_config('Time')
        if self.rec.channels <2:
            self.tfavg_rec.setDisabled(True)
        else:
            self.tfavg_rec.setEnabled(True)
            self.input_chan_box.clear()
            self.input_chan_box.addItems([str(i) for i in range(self.rec.channels)])

    def get_recording_mode(self):
        """
        Returns
        ----------
        str
            Current text of switch_rec_box
        """
        return self.switch_rec_box.currentText()

    def get_input_channel(self):
        """
        Returns
        ----------
        int
            Current index of input_chan_box
        """
        return self.input_chan_box.currentIndex()

    def get_record_config(self, *arg):
        """
        Returns
        ----------
        rec_configs: list
            List of recording settings
        """
        try:
            rec_configs = []
            data_type = [int,float,int,int,float]
            for cbox,dt in zip(self.rec_boxes,data_type):
                if type(cbox) == QComboBox:
                    #configs.append(cbox.currentText())
                    rec_configs.append(cbox.currentIndex())
                else:
                    config_input = cbox.text().strip(' ')
                    rec_configs.append(dt(float(config_input)))
            return(rec_configs)
        except Exception as e:
            print(e)
            return None

    def autoset_record_config(self, setting):
        """
        Recalculate samples or duration

        Parameters
        ----------
        setting: str
            Input type. Either 'Time' or 'Samples'
        """
        sample_validator = self.rec_boxes[0].validator()
        time_validator = self.rec_boxes[1].validator()

        if setting == "Time":
            valid = time_validator.validate(self.rec_boxes[1].text(),0)[0]
            if not valid == QValidator.Acceptable:
                self.rec_boxes[1].setText(str(time_validator.bottom()))

            samples = int(float(self.rec_boxes[1].text())*self.rec.rate)
            valid = sample_validator.validate(str(samples),0)[0]
            if not valid == QValidator.Acceptable:
                samples = sample_validator.top()
        elif setting == 'Samples':
            samples = int(self.rec_boxes[0].text())

        #samples = samples//self.rec.chunk_size  *self.rec.chunk_size
        duration = samples/self.rec.rate
        self.rec_boxes[0].setText(str(samples))
        self.rec_boxes[1].setText(str(duration))

    def reset_configs(self):
        """
        Reset the channels for triggering and reset validators
        """
        self.rec_boxes[3].clear()
        self.rec_boxes[3].addItems([str(i) for i in range(self.rec.channels)])

        validators = [QDoubleValidator(0.1,MAX_SAMPLE*self.rec.rate,1),
                     QIntValidator(-1,self.rec.chunk_size)]
        for cbox,vd in zip(self.rec_boxes[1:-2],validators):
            cbox.setValidator(vd)

    def update_TFavg_count(self,val):
        """
        Update the value of the number of recordings for average transfer function
        """
        self.avg_count_box.setText('Count: %i' % val)

    def toggle_trigger(self,string):
        """
        Enable or disable the trigger settings
        """
        try:
            val = int(string)
        except:
            val = -1

        if val == -1:
            self.rec_boxes[3].setEnabled(False)
            self.rec_boxes[4].setEnabled(False)
        else:
            self.rec_boxes[3].setEnabled(True)
            self.rec_boxes[4].setEnabled(True)

#-----------------------------????????------------------------------
class ChanLineText(QLineEdit):
    returnPressed = pyqtSignal(list)

    def __init__(self,*arg,**kwarg):
        super().__init__(*arg,**kwarg)

        self.invalidchar_re = re.compile(r'[^\d,:]+')

    def keyPressEvent(self,event):
        if event.key() == Qt.Key_Enter or event.key() == Qt.Key_Return:
            # REGEX here
            string = self.text()
            true_string = re.sub(self.invalidchar_re,'',string)
            self.setText(true_string)
            true_input = true_string.split(',')

            self.returnPressed.emit(true_input)

        else:
            QLineEdit.keyPressEvent(self,  event)

def set_input_limits(linebox,low,high,in_type):
    val = in_type(linebox.text())
    linebox.setText( str(min(max(val,low),high)) )


import weakref
import sys
if __name__ == '__main__':
    sys.path.append('../../')
import numpy as np
import pyqtgraph as pg
from pyqtgraph import ImageItem
from cued_datalogger.api.pyqt_extensions import matplotlib_lookup_table
from PyQt5.QtWidgets import(QWidget,QMenu,QAction,QActionGroup,QWidgetAction,QGridLayout,
                            QCheckBox,QRadioButton,QLineEdit,QSpinBox,QComboBox,
                            QLabel, QApplication, QVBoxLayout, QHBoxLayout, QPushButton)
from PyQt5.QtGui import QDoubleValidator
from PyQt5.QtCore import QMetaObject,QSize,QCoreApplication, QTimer, pyqtSignal


class InteractivePlotWidget(QWidget):
    """A QWidget containing a :class:`CustomPlotWidget` with mouse tracking
    crosshairs, a :class:`LinearRegionItem`, and spinboxes
    to display and control the values of the bounds of the linear region.
    Any additional arguments to :method:`__init__` are passed to the
    CustomPlotWidget.

    Attributes
    ----------
    PlotWidget : pg.PlotWidget
        The PlotWidget contained in the InteractivePlotWidget.
    ViewBox : pg.ViewBox
        The ViewBox contained in the InteractivePlotWidget.
    region : pg.LinearRegionItem
        The LinearRegionItem contained in the InteractivePlotWidget.
    vline : pg.InfiniteLine
        Vertical mouse-tracking line.
    hline : pg.InfiniteLine
        Horizontal mouse-tracking line.
    label : pg.LabelItem
        LabelItem displaying current mouse position.
    lower_box : QSpinBox
        QSpinBox displaying lower bound of :attr:`region`.
    upper_box : QSpinBox
        QSpinBox displaying upper bound of :attr:`region`.
    zoom_btn : QPushButton
        Press to zoom to the :attr:`region` with a set amount of padding.
    show_region : bool
        Controls whether the region is displayed.
    show_crosshair : bool
        Controls whether the crosshair is displayed.
    sig_region_changed : pyqtSignal([int, int])
        The signal emitted when the region is changed.
    """

    sig_region_changed = pyqtSignal([float, float])

    def __init__(self, parent=None,
                 show_region=True, show_crosshair=True, show_label=True,
                 *args, **kwargs):
        self.parent = parent

        super().__init__(parent)

        layout = QVBoxLayout(self)

        # # Set up the PlotWidget
        self.PlotWidget = CustomPlotWidget(self, show_region=show_region,
                                           show_crosshair=show_crosshair,
                                           show_label=show_label,
                                           *args, **kwargs)

        self.PlotWidget.region.sigRegionChanged.connect(self.updateBoxFromRegion)

        self.PlotItem = self.PlotWidget.PlotItem
        self.ViewBox = self.PlotWidget.ViewBox

        layout.addWidget(self.PlotWidget)

        # # Set up the region controls
        control_layout = QHBoxLayout()
        self.lower_box = pg.SpinBox(self, bounds=(None, None))
        self.lower_box.valueChanged.connect(self.updateRegionFromBox)
        self.lower_box_proxy = pg.SignalProxy(self.lower_box.valueChanged,
                                              rateLimit=30,
                                              slot=self.on_region_changed)
        self.upper_box = pg.SpinBox(self, bounds=(None, None))
        self.upper_box_proxy = pg.SignalProxy(self.upper_box.valueChanged,
                                              rateLimit=30,
                                              slot=self.on_region_changed)
        self.upper_box.valueChanged.connect(self.updateRegionFromBox)

        self.zoom_btn = QPushButton('Zoom', self)
        self.zoom_btn.clicked.connect(self.zoomToRegion)

        control_layout.addWidget(QLabel('Lower', self))
        control_layout.addWidget(self.lower_box)
        control_layout.addWidget(QLabel('Upper', self))
        control_layout.addWidget(self.upper_box)
        control_layout.addWidget(self.zoom_btn)
        layout.addLayout(control_layout)

        self.clear()

    def getPlotItem(self):
        return self.PlotItem

    def on_region_changed(self):
        lower = self.lower_box.value()
        upper = self.upper_box.value()
        self.sig_region_changed.emit(lower, upper)

    def clear(self):
        """Clear the PlotWidget and add the default items back in."""
        self.PlotWidget.clear()

    def updateRegionFromBox(self):
        # Get the bounds of the region as defined by the spinboxes
        region_bounds = [self.lower_box.value(),
                         self.upper_box.value()]
        # Sort them so that the lower one is first
        region_bounds.sort()
        # Update the region
        self.PlotWidget.region.setRegion(region_bounds)
        # Update the spinboxes
        self.lower_box.setValue(region_bounds[0])
        self.upper_box.setValue(region_bounds[1])

    def updateBoxFromRegion(self):
        # Get the region bounds as defined by the region
        region_bounds = list(self.PlotWidget.region.getRegion())
        # Sort it so the lowest is first
        region_bounds.sort()
        # Set the spinboxes to reflect the bounds of the region
        self.lower_box.setValue(region_bounds[0])
        self.upper_box.setValue(region_bounds[1])

    def zoomToRegion(self, padding=0.1):
        """Zoom to the region, with given padding."""
        lower, upper = self.PlotWidget.region.getRegion()
        self.PlotItem.setXRange(lower, upper, padding=padding)

    def getRegionBounds(self):
        """Return the lower and upper bounds of the region."""
        return self.lower_box.value(), self.upper_box.value()

    def plot(self, x=None, y=None, *args, **kwargs):
        """:func:`update_limits` from the x and y values, then plot
        the data on the plotWidget."""
        self.PlotWidget.plot(x, y, *args, **kwargs)

    def update_limits(self, x, y):
        """Set the increment of the spinboxes, the limits of zooming and
        scrolling the PlotItem, and move the region to x=0"""
        if x is not None and y is not None:
            # Update the increment of the spinboxes
            try:
                self.lower_box.setSingleStep(x.max()/100)
                self.upper_box.setSingleStep(x.max()/100)

                # Set the linear region to be in view
                #self.lower_box.setValue(x.max()*0.4)
                #self.upper_box.setValue(x.max()*0.6)
                #self.lower_box.setValue(0)
                #self.upper_box.setValue(0)

                # Set the limits of the PlotItem
                self.PlotItem.setLimits(xMin=x.min(), xMax=x.max())
                self.PlotItem.setRange(xRange=(x.min(), x.max()),
                                       yRange=(y.min(), y.max()),
                                       padding=0.2)
            except:
                print("Error in scaling limits.")


class CustomPlotWidget(pg.PlotWidget):
    def __init__(self, *args,
                 show_region=True, show_crosshair=True, show_label=True,
                 **kwargs):

        self.show_region = show_region
        self.show_crosshair = show_crosshair
        self.show_label = show_label

        super().__init__(*args, **kwargs)

        self.PlotItem = self.getPlotItem()
        self.PlotItem.disableAutoRange()
        self.ViewBox = self.PlotItem.getViewBox()

        #------------- Manually override wrapped methods ----------------------
        # Pyqtgraph's PlotWidget wraps some methods from PlotItem and ViewBox
        # In order to use our defined functions for these methods, we need
        # to override them using setattr rather than just redefining them
        methods_to_override = ["plot", "clear", "autoRange"]

        for method in methods_to_override:
            setattr(self, method, getattr(self, method + "_override"))

        #--------------------- Add default items ------------------------------
        # # Crosshair
        self.crosshair_x = pg.InfiniteLine(angle=0)
        self.crosshair_y = pg.InfiniteLine(angle=90)
        self.proxy = pg.SignalProxy(self.scene().sigMouseMoved, rateLimit=30,
                                    slot=self.mouseMoved)
        self.PlotItem.addItem(self.crosshair_y)
        self.PlotItem.addItem(self.crosshair_x)

        # # Region
        self.region = pg.LinearRegionItem(values=[0,0])
        self.PlotItem.addItem(self.region)

        # # Label
        self.label = pg.LabelItem(angle = 0)
        self.label.setParentItem(self.ViewBox)

        #---------------- Customise menus -------------------------------------
        # # Customise the plot item menu
        self.plotitem_menu = self.PlotItem.ctrlMenu
        self.plotitem_controls = self.PlotItem.ctrl

        for action in self.plotitem_menu.actions():
            # Remove all but two actions
            if action.text() not in ["Alpha", "Grid"]:
                self.plotitem_menu.removeAction(action)
            # Rename Alpha -> Transparency (easier to understand)
            if action.text() == "Alpha":
                action.setText("Transparency")
                self.plotitem_controls.alphaGroup.setTitle("Enable "
                                                           "transparency")

        self.show_crosshair_action = QAction("Show crosshair")
        self.show_crosshair_action.setCheckable(True)
        self.show_crosshair_action.triggered.connect(self.set_show_crosshair)
        self.show_crosshair_action.setChecked(self.show_crosshair)
        self.plotitem_menu.addAction(self.show_crosshair_action)

        self.show_region_action = QAction("Show region")
        self.show_region_action.setCheckable(True)
        self.show_region_action.triggered.connect(self.set_show_region)
        self.show_region_action.setChecked(self.show_region)
        self.plotitem_menu.addAction(self.show_region_action)

        self.show_label_action = QAction("Show Label")
        self.show_label_action.setCheckable(True)
        self.show_label_action.triggered.connect(self.set_show_label)
        self.show_label_action.setChecked(self.show_label)
        self.plotitem_menu.addAction(self.show_label_action)

        # # Customise the view box menu
        self.viewbox_menu = self.ViewBox.menu

        # Get rid of the current AutoRange action
        for action in self.viewbox_menu.actions():
            if action.text() == "View All":
                self.viewbox_menu.removeAction(action)

        # Add in our own AutoRange action
        self.autorange_action = QAction("Autorange", self)
        self.autorange_action.triggered.connect(self.autoRange)
        self.viewbox_menu.insertAction(self.viewbox_menu.actions()[0],
                                       self.autorange_action)

    #--------------------- Plot configuration ---------------------------------
    def mouseMoved(self, mouse_moved_event):
        """
        Update the crosshair and label to match the mouse position.
        """
        # Using a signal proxy turns the event into a tuple
        mouse_position = mouse_moved_event[0]

        # Check if the mouse is in the PlotItem
        if self.PlotItem.sceneBoundingRect().contains(mouse_position):
            # Convert it to the coordinate system
            try:
                mouse_point = self.ViewBox.mapSceneToView(mouse_position)
                # Update the label
                self.label.setText("<span style='font-size: 12pt;"
                                     "color: black'>"
                                     "x={:.2e},"
                                     "<span style='color: red'>"
                                     "y={:.2e}</span>".format(mouse_point.x(),
                                                              mouse_point.y()))
                self.crosshair_y.setPos(mouse_point.x())
                self.crosshair_x.setPos(mouse_point.y())

            except np.linalg.LinAlgError:
                print("Warning: Exception raised in calculating mouse position"
                      " (np.linalg.LinAlgError)")

    def autoRange_override(self, padding=None, items=None):
        """Autorange the view to fit the plotted data, ignoring the location
        of the crosshair. Accessed as :func:`autoRange`, not as
        :func:`autoRange_override`.
        """
        if items is None:
            items = self.PlotItem.items
        if self.crosshair_x in items:
            items.remove(self.crosshair_x)
        if self.crosshair_y in items:
            items.remove(self.crosshair_y)

        self.ViewBox.autoRange(padding=padding, items=items)

    def set_show_crosshair(self, show_crosshair):
        """Set whether the crosshair is visible."""
        self.show_crosshair = show_crosshair

        if self.show_crosshair:
            self.addItem(self.crosshair_x)
            self.addItem(self.crosshair_y)
        else:
            self.removeItem(self.crosshair_x)
            self.removeItem(self.crosshair_y)

    def set_show_region(self, show_region):
        """Set whether the region is visible."""
        self.show_region = show_region

        if self.show_region:
            self.addItem(self.region)
        else:
            self.removeItem(self.region)

    def set_show_label(self, show_label):
        """Set whether the label is visible."""
        self.show_label = show_label

        if self.show_label:
            self.label.show()
        else:
            self.label.hide()

    #-------------------- Plot interaction ------------------------------------
    def plot_override(self, *args, **kwargs):
        """Plot data on the widget and autoRange.
        Accessed as :func:`plot`, not as :func:`plot_override`."""

        self.PlotItem.plot(*args, **kwargs)
        self.autoRange(padding=0)

    def clear_override(self, *args, **kwargs):
        """Clear the PlotItem and add the default items back in.
        Accessed as :func:`clear`, not as :func:`clear_override`."""
        self.PlotItem.clear()
        self.set_show_crosshair(self.show_crosshair)
        self.set_show_region(self.show_region)


class ColorMapPlotWidget(InteractivePlotWidget):
    """An InteractivePlotWidget optimised for plotting color(heat) maps.
    Uses the Matplotlib colormap given by *cmap* to color the map.

    Attributes
    ----------
    lookup_table : ndarray
        The lookup table generated from *cmap* to colour the image with
    num_contours : int
        The number of different colour levels to plot
    contour_spacing : int
        How closely spaced the colour levels are
    """
    def __init__(self, parent=None, cmap="jet"):
        self.lookup_table = matplotlib_lookup_table(cmap)
        self.num_contours = 5
        self.contour_spacing_dB = 5
        self.parent = parent
        super().__init__(parent=self.parent)

    def plot_colormap(self, x, y, z, num_contours=5, contour_spacing_dB=5):
        """Plot *x*, *y* and *z* on a colourmap, with colour intervals defined
        by *num_contours* at *contour_spacing_dB* intervals."""

        #self.PlotWidget.removeItem(self.z_img)

        self.x = x
        self.y = y
        self.z = z

        self.num_contours = num_contours
        self.contour_spacing_dB = contour_spacing_dB
        self.update_lowest_contour()

        # Set up axes:
        x_axis = self.PlotWidget.getAxis('bottom')
        y_axis = self.PlotWidget.getAxis('left')

        self.x_scale_fact = self.get_scale_fact(x)
        self.y_scale_fact = self.get_scale_fact(y)

        x_axis.setScale(self.x_scale_fact)
        y_axis.setScale(self.y_scale_fact)

        #self.autoRange()

        self.z_img = ImageItem(z.transpose())
        self.z_img.setLookupTable(self.lookup_table)
        self.z_img.setLevels([self.lowest_contour, self.highest_contour])

        self.PlotWidget.addItem(self.z_img)

        self.PlotWidget.autoRange()
        #self.PlotWidget.ViewBox.autoRange()

    def get_scale_fact(self, var):
        return var.max() / var.size

    def update_lowest_contour(self):
        """Find the lowest contour to plot, as determined by the number of
        contours and the contour spacing."""
        self.lowest_contour = self.z.max() - (self.num_contours * self.contour_spacing_dB)
        self.highest_contour = self.z.max()


if __name__ == '__main__':

    pg.setConfigOption('background', 'w')
    pg.setConfigOption('foreground', 'k')
    pg.setConfigOption('antialias', True)
    defaultpen = pg.mkPen('k')

    app = 0
    app = QApplication(sys.argv)
    w = InteractivePlotWidget()
    x = np.linspace(0, 20*np.pi, 1e4)
    y = np.sin(x)
    w.plot(x, y, pen=defaultpen)
    w.show()

    sys.exit(app.exec_())



import numpy as np


def to_dB(x):
    """A simple function that converts x to dB: ``20*np.log10(x)``"""
    result = np.asarray(20*np.log10(x))
    result[~np.isfinite(result)] = 0
    return result


def from_dB(x):
    """A simple function that converts x in dB to a ratio over 1:
    ``10**(x/20)``

    """
    result = np.power(10, (x/20))
    return result


class MatlabList(list):
    """A list that allows slicing like Matlab.

    eg: ``l[1, 2, slice(3, 5), slice(10, 20, 2)]``

    """
    def __getitem__(self, index):
        output = []
        if isinstance(index, tuple) or isinstance(index, range):
            for i in index:
                if isinstance(i, range):
                    range_list = list(i)
                    for r in range_list:
                        if r not in index:
                            output.append(self[r])
                else:
                    output.append(self[i])
            return output
        else:
            return super().__getitem__(index)


def sdof_modal_peak(w, wr, zr, cr, phi):
    """Return a modal peak generated from the given parameters.

    Parameters
    ----------
    w : ndarray
        An array of omega (angular frequency) values.
    wr : float
        The resonant angular frequency.
    zr : float
        The damping factor.
    cr : float
        The magnitude of the modal constant.
    phi : float
        The phase of the modal constant.

    Returns
    -------
    ndarray
        The modal peak.

    Notes
    -----
    The modal peak is generated by:

    .. math:: \frac{C_r e^(i\phi)}{\omega_r^2 + 2i\zeta_r\omega_r^2 - \omega^2}

    """
    return cr*np.exp(1j*phi) / (wr**2 - w**2 + 2j * zr * wr**2)



# -*- coding: utf-8 -*-
"""
Created on Wed Aug  2 16:24:57 2017

@author: eyt21

This module contains the widget to open the window to edit metadata
in the acqusition window
"""
import sys,traceback
from PyQt5.QtWidgets import (QWidget,QVBoxLayout,QHBoxLayout,QMainWindow,
    QPushButton, QDesktopWidget,QStatusBar, QLabel,QLineEdit, QFormLayout,
    QGroupBox,QRadioButton,QSplitter,QFrame, QComboBox,QScrollArea,QGridLayout,
    QCheckBox,QButtonGroup,QTextEdit,QApplication,QListWidget,QDialog,QListWidgetItem )
from PyQt5.QtGui import QValidator,QIntValidator,QDoubleValidator,QColor,QPalette,QSizePolicy
from PyQt5.QtCore import Qt, QTimer, pyqtSignal, QPoint
import functools as fct

class ChanMetaWin(QDialog):
    """
    This is the Modal Dialog Window to edit metadata from acquisition window.
    it shows the channel names on the left in a list, and the metadata on the right.
    
    Attributes
    ----------
    livewin: acquisition window
        Window to get the metadata from
    all_info: list
        Contains metadata for each channel
    channel_listview: QListWidget
        Display list of names of channels
    meta_configs: list
        Widget for ('Channel', 'Name', 'Calibration Factor', 'Tags', 'Comments')
        of types   (QLabel, QLineEdit, QLineEdit, QLineEdit, CommentBox)
    """
    def __init__(self,livewin = None):
        """
        Parameters
        ----------
        livewin: acquisition window
            Window to get the metadata from
        """
        super().__init__()
        
        self.livewin = livewin
        
        self.setGeometry(200,500,500,500)
        self.setWindowTitle('Channels Info')
        self.setModal(True)
        
        self.initUI()
        self.setFocus()
        self.show()
        
    def initUI(self):
        """
        Initialise the UI
        """
        main_layout = QHBoxLayout(self)
        
        self.channel_listview = QListWidget(self)
        self.channel_listview.itemSelectionChanged.connect(self.display_metadata)

        self.all_info = []
        for i in range(len(self.livewin.live_chanset)):
                self.all_info.append(self.livewin.live_chanset.channel_metadata(i))
        
        for i in range(len(self.all_info)):
            self.channel_listview.addItem(self.all_info[i]['name'])

        
        main_layout.addWidget(self.channel_listview)
        
        channel_meta_form = QGridLayout()
        
        title = QLabel('CHANNEL METADATA',self)
        title.setAlignment(Qt.AlignCenter)
        channel_meta_form.addWidget(title,0,0,1,3)
        meta_dname = ('Channel','Name','Calibration Factor','Tags','Comments')
        meta_dtype = ('','name','calibration_factor','tags','comments')
        
        self.meta_configs = []
        UI_type = (QLabel,QLineEdit,QLineEdit,QLineEdit,CommentBox)
        
        row = 1
        for m,UI,md in zip(meta_dname,UI_type,meta_dtype):
            channel_meta_form.addWidget(QLabel(m,self),row,0,1,1)
            cbox = UI(self)
            if not UI == QLabel:
                cbox.editingFinished.connect(fct.partial(self.update_metadata,md,cbox))
            channel_meta_form.addWidget(cbox,row,1,1,2)
            self.meta_configs.append(cbox)
            row +=1
        self.channel_listview.setCurrentRow(0)
        row +=5
        
        apply_btn = QPushButton('Apply',self)
        yes_btn = QPushButton('Ok',self)
        yes_btn.clicked.connect(self.export_metadata)
        no_btn =  QPushButton('Cancel',self)
        no_btn.clicked.connect(self.closing)
        channel_meta_form.addWidget(apply_btn,row,0)
        channel_meta_form.addWidget(yes_btn,row,1)
        channel_meta_form.addWidget(no_btn,row,2)
        
        main_layout.addLayout(channel_meta_form)
        
    def display_metadata(self):
        """
        Display the selected channel metadata
        """
        sel_num = self.channel_listview.currentRow()
        self.meta_configs[0].setText(str(sel_num))
        meta_dtype = ('name','calibration_factor','tags','comments')
        
        for n,md in enumerate(meta_dtype,1):
            self.meta_configs[n].setText(str(self.all_info[sel_num][md]))
            
    def update_metadata(self,meta_name,UI):
        """
        Update the selected channel metadata
        """
        chan_num = self.channel_listview.currentRow()
        string = UI.text()
        try:
            if meta_name == 'tags':
                string = string.split(' ')
            self.all_info[chan_num][meta_name] = string
            
            if meta_name == 'name':
                self.channel_listview.currentItem().setText(string)
        except:
            t,v,tb = sys.exc_info()
            print(t)
            print(v)
            print(traceback.format_tb(tb))
            
    def export_metadata(self):
        """
        Export the metadata to the livewin ChannelSet
        """
        try:
            for i in range(len(self.all_info)):
                self.livewin.live_chanset.set_channel_metadata(i,self.all_info[i])
        except:
            t,v,tb = sys.exc_info()
            print(t)
            print(v)
            print(traceback.format_tb(tb))
            
        self.closing()
        
    def closing(self):
        self.done(0)
        
        
class CommentBox(QTextEdit):
    """
    Reimplement QTextEdit to be similar to QLineEdit,
    i.e. having editingFinished signal and text()
    """
    editingFinished = pyqtSignal()
        
    def focusOutEvent(self,event):
        self.editingFinished.emit()
        QTextEdit.focusOutEvent(self,event)
        
    def text(self):
        return self.toPlainText()
 
if __name__ == '__main__':
    app = 0 
    app = QApplication(sys.argv)
    w = ChanMetaWin()
    sys.exit(app.exec_())


import sys
from PyQt5.QtWidgets import QApplication

if __name__ == '__main__':
    sys.path.append('../')

import cued_datalogger
from cued_datalogger.api.workspace import Workspace
from cued_datalogger.analysis.analysis_window import AnalysisWindow
from cued_datalogger import __version__
import argparse

def run():
    parser = argparse.ArgumentParser(description="The CUED DataLogger (v{}) "
                                     "for acquiring and analysing data. \n"
                                     "See cued-datalogger.readthedocs.io for "
                                     "documentation.".format(__version__))
    parser.add_argument('-w', '--workspace', dest='workspace_path',
                        help="Run the datalogger with the configuration given "
                        "in the Workspace (.wsp) file found at WORKSPACE_PATH")

    args = parser.parse_args()

    print("CUED DataLogger {}".format(__version__))

    # Create the application instance
    app = 0
    app = QApplication(sys.argv)

    # Create the workspace instance
    CurrentWorkspace = Workspace()

    if args.workspace_path is not None:
        CurrentWorkspace.load(args.workspace_path)

    # Create the window
    w = AnalysisWindow()

    # Connect the window to the workspace
    CurrentWorkspace.set_parent_window(w)

    w.addon_widget.discover_addons(w.CurrentWorkspace.path + "/addons/")

    # Run the program
    w.show()
    sys.exit(app.exec_())


if __name__ == '__main__':
    run()



from cued_datalogger import api
from cued_datalogger import analysis
from cued_datalogger import acquisition
from cued_datalogger.analysis import analysis_window as analysis_window
from cued_datalogger.acquisition import acquisition_window as acquisition_window
from cued_datalogger.api import workspace as workspace

import os.path as _path

_PKG_ROOT = _path.abspath(_path.dirname(__file__))

if _path.isfile(_path.join(_PKG_ROOT, 'VERSION')):
    with open(_path.join(_PKG_ROOT, 'VERSION')) as _version_file:
    #with open('VERSION') as _version_file:
        __version__ = _version_file.read()
else:
    __version__ = None



# -*- coding: utf-8 -*-
#Created on Sat Jul 15 20:01:58 2017
#@author: En Yi
"""
This module contains the abstract class to implement a proper    
Recorder Class. To do so, subclass RecorderParent when creating
a new Recorder class.

Example:
    from RecorderParent import RecorderParent
    
    class newRecorder(RecorderParent):
        ...
        ...
        
If you have PyQt, it will import RecEmitter for emitting Signals.

Attributes
----------
    QT_EMITTER : Indicates whether you can use qt Signals

"""
from abc import ABCMeta, abstractmethod
import numpy as np
import copy as cp

try:
    from cued_datalogger.acquisition.RecEmitter import RecEmitter
    QT_EMITTER = True
except Exception as e:
    print(e)
    QT_EMITTER = False
    


class RecorderParent(object):
    """
    Recorder abstract class. Sets up the buffer and skeleton for audio streaming
    
    Attributes
    ----------
    channels: int
        Number of Channels
    rate: int
        Sampling rate
    chunk_size: int
        Number of samples to get from each channel in one chunk
    num_chunk: int
        Number of chunks to store in circular buffer
    recording: bool
        Indicate whether to record
    """
    __metaclass__ = ABCMeta
    
#---------------- INITIALISATION METHODS -----------------------------------    
    def __init__(self,channels = 1,rate = 44100, chunk_size = 1024,
                 num_chunk = 4):
        """
        Initialise a ciruclar buffer, array and trigger for recording
        
        Parameters
        ----------
        channels: int
                Number of Channels
        rate: int
            Sampling rate
        chunk_size: int
            Number of samples to get from each channel in one chunk
        num_chunk: int
            Number of chunks to store in circular buffer
        """
        self.channels = channels
        self.rate = rate
        self.chunk_size = chunk_size
        self.num_chunk = num_chunk;
        self.audio_stream = None #: The audio object
        
        self.allocate_buffer()
        self.show_stream_settings()
        
        self.trigger_init()
        
        if QT_EMITTER:
            self.rEmitter = RecEmitter()
        else:
            self.rEmitter = None
                 
    def allocate_buffer(self):
        """
        Set up the circular buffer
        """
        self.buffer = np.zeros(shape = (self.num_chunk,
                                        self.chunk_size,
                                        self.channels))
        self.next_chunk = 0;

#---------------- DESTRUCTOR METHODS -----------------------------------     
    def __del__(self):
        self.close()
           
    def close(self):
        """
        Close the audio object, to be called if streaming is no longer needed 
        """
        self.stream_close()
 
#---------------- DEVICE SETTINGS METHODS -----------------------------------     
    def show_stream_settings(self):
        """
        Show the settings of the recorder
        """
        print('Channels: %i' % self.channels)
        print('Rate: %i' % self.rate)
        print('Chunk size: %i' % self.chunk_size)
        print('Number of chunks: %i' % self.num_chunk)
        
    def set_filename(self,filename):
        self.filename = filename
        
    @abstractmethod
    def set_device_by_name(self, name):
        """
        Set the device to be used for audio streaming
        """
        pass
    
    @abstractmethod    
    def available_devices(self):
        """
        Displays all available device for streaming
        """
        pass
    
    @abstractmethod
    def current_device_info(self):
        """
        Displays information about available the current device set
        """
        pass
    
#---------------- DATA METHODS -----------------------------------
    def audiodata_to_array(self,data):
        """
        Convert audio data obtained into a proper array
        
        Parameters
        -----------
        data: Numpy Array
            Audio data 
        """
        return data.reshape((-1,self.channels))/ 2**15
    
#---------------- BUFFER METHODS -----------------------------------
    def write_buffer(self,data):
        """
        Write the data obtained into buffer and move to the next chunk
        
        Parameters
        -----------
        data: Numpy Array
            Audio data 
        """
        self.buffer[self.next_chunk,:,:] = data
        self.next_chunk = (self.next_chunk + 1) % self.num_chunk
     
    def get_buffer(self):
        """
        Convert the buffer data as a 2D array by stitching the chunks together
        
        Returns
        ----------
        Buffer data: Numpy Array
            with dimension of(chunk_size * num_chunk) x channels
            The newest data on the most right 
        """
        return np.concatenate((self.buffer[self.next_chunk:],self.buffer[:self.next_chunk]),axis = 0) \
                 .reshape((self.buffer.shape[0] * self.buffer.shape[1],
                           self.buffer.shape[2]))
        
#---------------- RECORDING METHODS -----------------------------------
    def open_recorder(self):
        """
        Initialise the variables for recording.
        """
        self.recording = False
        self.initialised_record = False
        self.next_rec_chunk = 0
        self.total_rec_chunk = 0
        self.recorded_data = []
        
    def record_init(self,samples = None,duration = 3):
        """
        Remove any pretrigger and postrigger data
        Calculate the number of chunk to record
        It will record more samples than necessary, then slice down to the
        amount of samples required + putting in pretrigger data
        
        Parameters
        ----------
        samples: int
            Number of samples to record
        duration: int
            The recording duration
        """
        if not self._record_check():
            return False
        
        self.pretrig_data = np.array([],dtype = np.int16)
        self.part_posttrig_data = np.array([],dtype = np.int16)
        if samples:
            self.actual_rec_samples = samples
            self.total_rec_chunk = (samples // self.chunk_size)+1
            self.rec_samples = self.total_rec_chunk * self.chunk_size 
        else:
            self.total_rec_chunk = int(duration * self.rate // self.chunk_size)
            self.rec_samples = self.total_rec_chunk * self.chunk_size 
            self.actual_rec_samples = self.rec_samples
            
        self.next_rec_chunk = 0
        
        self.initialised_record = True
        
        print('Recording function is ready! Use record_start() to start')
        
        return True
    
    def _record_check(self):
        """
        Check if it is possible to start a recording
        
        Returns
        ----------
        bool
            True if possible, False otherwise
        """
        if not self.audio_stream:
            print('No recording stream initiated!')
            return False
        
        # Check if the previous recorded data is flushed
        if self.recorded_data:
            print('Please flush your recorded data')
            return False
        
        return True

    # Function to initiate a normal recording
    def record_start(self):
        """
        Start recording if it is possible
        
        Returns
        ----------
        bool
            True if possible, False otherwise
        """
        if not self._record_check():
            return False
                       
        # Start the recording if recording is initialised
        if self.initialised_record: 
            self.stream_start()
            self.recording = True
            print('Recording Start!')
            return True
        else:
            print('Record not initialised! Use record_init(duration) first!')
            return False
        
    def _record_stop(self):
        """
        Stop a successful recording and emit a signal if possible
        """
        # Stop the recording
        self.recording = False
        # Give a signal that recording is done
        print('Recording Done! Please flush the data with flush_record_data().')
        if self.rEmitter:
            self.rEmitter.recorddone.emit()
            
    def record_cancel(self):
        """
        Cancel a recording and clear any recorder data
        """
        print('Recording Cancel! Recorded data has been discarded!')
        self.trigger = False
        self.recording = False
        self.recorded_data = []
           
    def record_data(self,data):
        """
        Append recorded chunk to recorder_data
        and stop doing so if neccessary amount of chunks is recorded
        """
        #data = cp.copy(self.buffer[self.next_chunk-1])
        self.recorded_data.append(data)
        # Check to see whether recording is done
        self.next_rec_chunk += 1
        if self.next_rec_chunk == self.total_rec_chunk:
            self._record_stop()
          
    def flush_record_data(self):
        """
        Add in any partial posttrigger data
        Slice the recorded data into the requested amount of samples
        Add in any pretrigger data
        
        Returns
        ----------
        flushed_data: numpy array
            2D numpy array (similar to get_buffer) 
        """
        if self.recorded_data:
            data =  np.array(self.recorded_data);
            flushed_data = data.reshape((self.rec_samples,self.channels))
            if self.part_posttrig_data.shape[0]:
                flushed_data = np.vstack((self.part_posttrig_data,flushed_data))
            
            flushed_data = flushed_data[:self.actual_rec_samples,:]

            self.recorded_data = []
            if self.pretrig_data.shape[0]:
                flushed_data = np.vstack((self.pretrig_data,flushed_data))
                
            print('Data flushed')
            return flushed_data               
                            

#---------------- STREAMING METHODS ----------------------------------- 
    @abstractmethod                                    
    def stream_init(self, playback = False):
        """
        Callback function for initialising audio streaming.
                
        Parameters
        ----------
        playback: bool
            Whether to output the stream to a device
            
        Returns
        ----------
        bool
            True if successful, False otherwise
        """
        pass
            
    @abstractmethod 
    def stream_start(self):
        """
        Callback function for starting the audio streaming.
        """
        pass
    @abstractmethod 
    def stream_stop(self):
        """
        Callback function for stopping the audio streaming.
        """
        pass
        
    @abstractmethod 
    def stream_close(self):
        """
        Callback function for closing the audio streaming.
        """
        pass
#---------------- TRIGGER METHODS -----------------------------------
    def trigger_init(self):
        """
        Initialise the variable for the trigger recording
        """
        self.trigger = False
        self.trigger_threshold = 0
        self.trigger_channel = 0
        self.ref_level = 0.08
        self.pretrig_samples = 200
        self.pretrig_data = np.array([])
        self.part_posttrig_data = np.array([])
        
    def trigger_start(self,duration = 3, threshold = 0.09, channel = 0,pretrig = 200,posttrig = 5000):
        """
        Start the trigger if possible
        
        Returns
        ----------
        bool
            True if successful, False otherwise
        """
        if self.recording:
            print('You are current recording. Please finish the recording before starting the trigger.')
            return False
        
        if not self.trigger:
            if not self.record_init(samples = posttrig):
                return False
            self.trigger = True
            self.trigger_threshold = threshold
            self.trigger_channel = channel
            self.pretrig_samples = pretrig
            self.ref_level = np.mean(self.buffer[self.next_chunk,:,self.trigger_channel])
            print('Reference level: %.2f' % self.ref_level)
            print('Trigger Set!')
            return True
        else:
            print('You have already started a trigger')
            return False

    def _trigger_check_threshold(self,data):
        """
        Check if the trigger is set off
        Start recording if so and emit a signal if possible
        
        Parameters
        ----------
        data: Numpy Array
            data to be analysed
        """
        trig_data = data[:,self.trigger_channel]
        norm_data = abs(trig_data - np.mean(trig_data))#- self.ref_level
        
        maximum = np.amax(norm_data)
        
        if maximum > self.trigger_threshold:
            print('Triggered!')
            pos = np.argmax(norm_data)
            self.recording = True
            self.trigger = False
            try:
                self.part_posttrig_data =  cp.copy(self.buffer[self.next_chunk-1, pos:,:])
                temp = np.vstack((self.buffer[self.next_chunk-2,:,:],self.buffer[self.next_chunk-1,:pos,:]))
                self.pretrig_data = temp[temp.shape[0]-self.pretrig_samples:,:]
            except Exception as e:
                print(e)
                print('Cannot get trigger data')
            if self.rEmitter:
                self.rEmitter.triggered.emit()
            
#----------------- DECORATOR METHODS --------------------------------------
    @property
    def num_chunk(self):
        """
        int 
            Number of chunks to store in circular buffer
            The setter method will calculate the maximum possible number of chunks
            based on an arbitrary number of sample limit (2^25 in here)
        """
        return self._num_chunk

    @num_chunk.setter
    def num_chunk(self, num_chunks):
        n = max(1, int(num_chunks))
        try:
            if n * self.chunk_size > 2**25:
                n = 2**16 // self.chunk_size
            self._num_chunk = n
            self.allocate_buffer()
            #print(self._num_chunk)
        except Exception as e:
            #print(e)
            self._num_chunks = n
        
        
    @property
    def chunk_size(self):
        """
        int 
            Number of samples to get from each channel in one chunk
            The setter method will calculate the maximum possible size
            based on an arbitrary number of sample limit (2^25 in here)
        """
        return self._chunk_size

    @chunk_size.setter
    def chunk_size(self, chunk_size):
        n = max(1, int(chunk_size))
        try:
            if n * self.num_chunk > 2**25:
                n = 2**16 // self.num_chunk
            self._chunk_size = n
            #print(self._chunk_size)
            self.allocate_buffer()
        except Exception as e:
            #print(e)
            self._chunk_size = n

            
                                  
 


from PyQt5.QtCore import pyqtSignal, Qt
import numpy as np

# ------------------------------------------------------
# Basic PyQt Widgets
# ------------------------------------------------------
from PyQt5.QtWidgets import QSlider, QSpinBox, QVBoxLayout, QHBoxLayout, QWidget

class BaseNControl(QWidget):
    """
    A QWidget containing a paired SpinBox and Slider that are constrained to 
    show values in a given base.
    
    Attributes
    ----------
    valueChanged : pyqtSignal(int)
        The signal emitted when the current value is changed, containing the 
        current value.
    orientation : Qt.Vertical or Qt.Horizontal
        The orientation of the SpinBox - Slider pair
    base : int
        The current base that the BaseNControl is using.
    """
    valueChanged = pyqtSignal(int)
    
    def __init__(self, orientation, parent=None, 
                 show_spinbox=True, 
                 show_slider=True,
                 base=2):
        
        super().__init__(parent)
        
        self.parent = parent
        self.orientation = orientation
        self.base = base
        
        self.spinbox = BaseNSpinBox(self)
        self.slider = QSlider(self.orientation, self)
        
        # Update whenever enter is pressed in the spinbox
        self.spinbox.editingFinished.connect(self.on_spinbox_edit_finished)
        # Update when the buttons are pressed
        self.spinbox.sig_step_down.connect(self.decrease_value)
        self.spinbox.sig_step_up.connect(self.increase_value)
        # Update whenever the slider is moved
        self.slider.valueChanged.connect(self.set_n)
        #self.slider.actionTriggered.connect(self.on_slider_value_changed)
        
        if self.orientation == Qt.Vertical:
            layout = QVBoxLayout()
        elif self.orientation == Qt.Horizontal:
            layout = QHBoxLayout()
        else:
            raise ValueError("Unrecognised orientation (accepted values: "
                "Qt.Horizontal, Qt.Vertical)")
        
        if show_spinbox:
            layout.addWidget(self.spinbox)
        if show_slider:
            layout.addWidget(self.slider)
        
        self.setLayout(layout)
    
    def set_power_range(self, min_power_n, max_power_n):    
        """Set the range of allowed values to be from ``base**min_power_n`` to
        ``base**max_power_n``."""
        self.slider.setRange(min_power_n, max_power_n)
        self.spinbox.setRange(self.base**min_power_n, self.base**max_power_n)
        
        self.set_n(min_power_n)
    
    def set_value_range(self, min_value, max_value):
        """Set the range of allowed values to be from *min_value* to 
        *max_value*."""
        self.spinbox.setRange(min_value, max_value)
        min_n = int(np.floor(np.log(min_value)/np.log(self.base)))
        max_n = int(np.ceil(np.log(max_value)/np.log(self.base)))
        self.slider.setRange(min_n, max_n)
        
        self.set_n(min_n)

    def set_n(self, n, set_both=False):
        """If *n* is in range, set it as the current n-value.
        
        If *set_both*, set the values for both the slider and spinbox, 
        otherwise set the value for whichever did not send the signal."""
        if n >= self.slider.minimum() and n <= self.slider.maximum(): 
            self.n = n
            if set_both:
                self.spinbox.setValue(self.base**self.n)
                self.slider.setValue(self.n)
            elif self.sender() == self.slider:
                self.spinbox.setValue(self.base**self.n)
            elif self.sender() == self.spinbox:
                self.slider.setValue(self.n)     
            self.valueChanged.emit(self.base**self.n)

    def set_value(self, value):
        """Set the value for the slider and the spinbox to *value*."""
        n = int(np.round(np.log(value)/np.log(self.base)))
        self.set_n(n, True)
        
    def increase_value(self):
        self.set_n(self.n + 1, True)
        
    def decrease_value(self):
        self.set_n(self.n - 1, True)
    
    def on_spinbox_edit_finished(self):
        self.set_value(self.spinbox.value())
    
    def value(self):
        """Return the current value of the BaseNControl."""
        return self.spinbox.value()
        

class BaseNSpinBox(QSpinBox):
    
    sig_step_up = pyqtSignal()
    sig_step_down = pyqtSignal()
    
    def __init__(self, parent=None):
        super().__init__(parent)
    
    def stepBy(self, number_of_steps):
        if number_of_steps > 0:
            self.sig_step_up.emit()
        elif number_of_steps < 0:
            self.sig_step_down.emit()

# -----------------------------------------
# Matplotlib widgets
# -----------------------------------------
import matplotlib
from matplotlib.figure import Figure
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg as FigureCanvas

from PyQt5.QtWidgets import QSizePolicy


class MatplotlibCanvas(FigureCanvas):
    """A custom version of Matplotlib's :class:`FigureCanvas`, with a title, no
    toolbar, and Qt resizable functionality."""
    def __init__(self, suptitle=None):
        matplotlib.use('Qt5Agg')  
        matplotlib.rcParams['toolbar'] = 'None'

        fig = Figure()
        fig.suptitle(suptitle)
        self.axes = fig.add_subplot(111)

        self.init_plot()
        
        FigureCanvas.__init__(self, fig)
        FigureCanvas.setSizePolicy(self, QSizePolicy.Expanding, QSizePolicy.Expanding)
        FigureCanvas.updateGeometry(self)

    def init_plot(self):
        pass
    
    def draw_plot(self):
        pass
    
    def update_plot(self):
        pass

from matplotlib.cm import get_cmap

def matplotlib_lookup_table(name):
    """Return a pyqtgraph-style lookup table (*ndarray*) from a 
    :class:`~/matplotlib.colors.Colormap`"""
    cmap = get_cmap(name)
    return cmap(np.arange(256)) * 255


from cued_datalogger.api.pyqtgraph_extensions import InteractivePlotWidget
from cued_datalogger.api.toolbox import Toolbox
from cued_datalogger.api.numpy_extensions import to_dB
from cued_datalogger.api.channel import Channel

from PyQt5.QtWidgets import (QWidget, QGridLayout, QPushButton, QComboBox,
                             QCheckBox, QLabel, QGroupBox)
from PyQt5.QtCore import pyqtSignal

import numpy as np
from numpy.fft import rfft


class FrequencyDomainWidget(InteractivePlotWidget):
    """
    The FrequencyDomainWidget is the main display widget for everything in
    the frequency domain.

    Attributes
    ----------
    channels : list of Channel
        The currently selected channel objects
    current_plot_type : str
        Any of 'linear magnitude', 'log magnitude', 'phase', 'real part',
        'imaginary part', 'nyquist'. The current type of plot that is
        displayed.
    show_coherence : bool
        If `True`, coherence is also plotted on the axes.
    """
    def __init__(self, parent=None):
        super().__init__(parent)

        self.channels = []

        self.plot_types = ['linear magnitude',
                           'log magnitude',
                           'phase',
                           'real part',
                           'imaginary part',
                           'nyquist']
        self.current_plot_type = self.plot_types[0]

        self.plot_transfer_function = False
        self.show_coherence = False

    def set_selected_channels(self, selected_channels):
        """Update which channels are plotted. Sets `self.channels` to
        *selected_channels*."""
        # If no channel list is given
        if not selected_channels:
            self.channels = []
        else:
            self.channels = selected_channels

        self.update_plot(self.plot_transfer_function)

    def set_plot_type(self, plot_type):
        """Set what type of plot is displayed. *plot_type* can be any of
        'linear magnitude', 'log magnitude', 'phase', 'real part', 'imaginary
        part', 'nyquist'."""
        self.current_plot_type = plot_type.lower()
        self.update_plot(self.plot_transfer_function)

    def set_show_coherence(self, show_coherence):
        """Set whether the coherence is displayed."""
        self.show_coherence = show_coherence
        self.update_plot(plot_transfer_function=True)

    def update_plot(self, plot_transfer_function=False):
        """If *plot_transfer_function*, plot the transfer function. Otherwise,
        plot the spectrum."""
        self.plot_transfer_function = plot_transfer_function

        # Clear the plot
        self.clear()

        for channel in self.channels:
            # Extract the data
            if self.plot_transfer_function:
                if channel.is_dataset("transfer_function"):
                    data = channel.data("transfer_function")
                else:
                    print("{}: no 'transfer_function' "
                          "dataset.".format(channel.name))
                    continue
            else:
                if channel.is_dataset("spectrum"):
                    data = channel.data("spectrum")
                else:
                    print("{}: no 'spectrum' "
                          "dataset.".format(channel.name))
                    continue

            # Plot the data
            if self.current_plot_type == 'linear magnitude':
                self.plot(channel.data("frequency"),
                          np.abs(data),
                          pen=channel.colour)

            elif self.current_plot_type == 'log magnitude':
                self.plot(channel.data("frequency"),
                          to_dB(np.abs(data)),
                          pen=channel.colour)

            elif self.current_plot_type == 'phase':
                self.plot(channel.data("frequency"),
                          np.angle(data, deg=True),
                          pen=channel.colour)

            elif self.current_plot_type == 'real part':
                self.plot(channel.data("frequency"),
                          np.real(data),
                          pen=channel.colour)

            elif self.current_plot_type == 'imaginary part':
                self.plot(channel.data("frequency"),
                          np.imag(data),
                          pen=channel.colour)

            elif self.current_plot_type == 'nyquist':
                self.plot(np.real(data),
                          np.imag(data),
                          pen=channel.colour)

            if self.plot_transfer_function and self.show_coherence:
                print("Plotting coherence...")
                if channel.is_dataset("coherence"):
                    if self.current_plot_type == 'linear magnitude':
                        self.plot(channel.data("frequency"),
                                  np.abs(channel.data("coherence")))
                    elif self.current_plot_type == 'log magnitude':
                        self.plot(channel.data("frequency"),
                                  to_dB(np.abs(channel.data("coherence"))))
                    elif self.current_plot_type == 'phase':
                        self.plot(channel.data("frequency"),
                                  np.angle(channel.data("coherence"), deg=True))
                    elif self.current_plot_type == 'real part':
                        self.plot(channel.data("frequency"),
                                  channel.data("coherence").real)
                    elif self.current_plot_type == 'imaginary part':
                        self.plot(channel.data("frequency"),
                                  channel.data("coherence").imag)
                    elif self.current_plot_type == 'nyquist':
                        self.plot(channel.data("coherence").real,
                                  channel.data("coherence").imag)
                    print("Done.")
                else:
                    print("{}: no 'coherence' dataset".format(channel.name))

    def calculate_spectrum(self):
        """Calculate the frequency spectrum of all the selected channels."""
        print("Calculating spectrum...")
        for channel in self.channels:
            if channel.is_dataset("time_series"):
                time_series = channel.data("time_series")
                window = np.hanning(time_series.size)
                spectrum = rfft(time_series * window)

                if not channel.is_dataset("spectrum"):
                    channel.add_dataset("spectrum", data=spectrum)
                else:
                    channel.set_data("spectrum", spectrum)

            else:
                print("Skipping {}: no 'time_series' "
                      "dataset.".format(channel.name))
                continue
        print("Done.")
        self.update_plot()

    def calculate_transfer_function(self, input_channel=None):
        """Calculate the transfer function, using the channel object given by
        *input_channel* as the input. If no channel specified, treat the first
        selected channel as input."""

        if not input_channel:
            input_channel = self.channels[0]
        if not isinstance(input_channel, Channel):
            print("Error in calculating transfer function: no input channel.")
            pass

        print("Calculating transfer function...")

        input_spectrum = input_channel.data("spectrum")
        input_auto_spectrum = calculate_auto_spectrum(input_spectrum)

        for channel in self.channels:
            # Skip the input channel
            if channel == input_channel:
                pass

            if channel.is_dataset("spectrum"):
                # Get the data
                output_spectrum = channel.data("spectrum")

                # Calculate the auto- and cross- spectra
                output_auto_spectrum = \
                    calculate_auto_spectrum(output_spectrum)

                cross_spectrum = \
                    calculate_cross_spectrum(input_spectrum,
                                                  output_spectrum)

                # Calculate the transfer function
                transfer_function = output_auto_spectrum / cross_spectrum
                # Update the channel
                if not channel.is_dataset("transfer_function"):
                    channel.add_dataset("transfer_function", data=transfer_function)
                else:
                    channel.set_data("transfer_function", transfer_function)

                # Calculate the coherence
                coherence = (cross_spectrum * cross_spectrum.conj() /
                             input_auto_spectrum * output_auto_spectrum).real
                # Update the channel
                if not channel.is_dataset("coherence"):
                    channel.add_dataset("coherence", data=coherence)
                else:
                    channel.set_data("coherence", coherence)
            else:
                print("Skipping {}: no 'spectrum' "
                      "dataset.".format(channel.name))
                continue
        print("Done.")
        self.update_plot(plot_transfer_function=True)


def calculate_auto_spectrum(spectrum):
    return spectrum * spectrum.conj()

def calculate_cross_spectrum(input_spectrum, output_spectrum):
    return input_spectrum.conj() * output_spectrum

def compute_transfer_function(autospec_in,autospec_out,crossspec):
    transfer_func = (autospec_out/crossspec)
    coherence = ((crossspec * np.conjugate(crossspec))/(autospec_in*autospec_out))
    return(transfer_func,np.real(coherence))


class FrequencyToolbox(Toolbox):
    """Toolbox containing the Frequency Domain controls."""
    sig_convert_to_circle_fit = pyqtSignal()
    sig_plot_type_changed = pyqtSignal(str)
    sig_plot_transfer_function = pyqtSignal()
    sig_plot_frequency_spectrum = pyqtSignal()
    sig_show_coherence = pyqtSignal(bool)
    sig_calculate_transfer_function = pyqtSignal()

    def __init__(self, parent=None):
        super().__init__(parent=parent)
        self.parent = parent

        self.init_ui()

    def init_ui(self):
        # # Plot options tab
        self.plot_options_tab = QWidget(self)
        plot_options_tab_layout = QGridLayout()

        plot_options_tab_layout.addWidget(QLabel("Plot:"), 0, 0)
        self.current_plot_combobox = QComboBox(self)
        self.current_plot_combobox.addItems(['Frequency spectrum', 'Transfer function'])
        self.current_plot_combobox.currentIndexChanged[str].connect(self.on_current_plot_changed)
        plot_options_tab_layout.addWidget(self.current_plot_combobox, 1, 0)

        self.show_coherence_checkbox = QCheckBox('Show coherence',self)
        self.show_coherence_checkbox.stateChanged.connect(self.sig_show_coherence.emit)
        self.show_coherence_checkbox.stateChanged.connect(self.set_plot_transfer_function)
        plot_options_tab_layout.addWidget(self.show_coherence_checkbox, 2, 0)

        plot_options_tab_layout.addWidget(QLabel("Plot type:"), 3, 0)
        self.plot_type_combobox = QComboBox(self.plot_options_tab)
        self.plot_type_combobox.addItems(['Linear Magnitude',
                                          'Log Magnitude',
                                          'Phase',
                                          'Real Part',
                                          'Imaginary Part',
                                          'Nyquist'])
        self.plot_type_combobox.setCurrentIndex(0)
        self.plot_type_combobox.currentIndexChanged[str].connect(self.sig_plot_type_changed.emit)
        plot_options_tab_layout.addWidget(self.plot_type_combobox, 4, 0)

        plot_options_tab_layout.setRowStretch(5, 1)
        self.plot_options_tab.setLayout(plot_options_tab_layout)

        self.addTab(self.plot_options_tab, "Plot Options")

        # # Conversion tab
        self.convert_tab = QWidget(self)
        convert_tab_layout = QGridLayout()

        transfer_function_groupbox = QGroupBox("Transfer function")
        transfer_function_groupbox_layout = QGridLayout()

        label = QLabel("Use first selected channel as input to compute "
                       "transfer function")
        label.setWordWrap(True)
        transfer_function_groupbox_layout.addWidget(label, 0, 0)
        self.convert_to_transfer_function_button = \
            QPushButton("Compute transfer function")
        self.convert_to_transfer_function_button.clicked.connect(self.sig_calculate_transfer_function.emit)
        self.convert_to_transfer_function_button.clicked.connect(self.set_plot_transfer_function)
        transfer_function_groupbox_layout.addWidget(self.convert_to_transfer_function_button, 1, 0)

        transfer_function_groupbox.setLayout(transfer_function_groupbox_layout)
        convert_tab_layout.addWidget(transfer_function_groupbox, 0, 0)

        modal_fitting_groupbox = QGroupBox("Modal fitting")
        modal_fitting_groupbox_layout = QGridLayout()

        self.circle_fit_btn = QPushButton("Circle fit")
        self.circle_fit_btn.clicked.connect(self.sig_convert_to_circle_fit.emit)
        modal_fitting_groupbox_layout.addWidget(self.circle_fit_btn, 1, 0)

        modal_fitting_groupbox.setLayout(modal_fitting_groupbox_layout)
        convert_tab_layout.addWidget(modal_fitting_groupbox, 1, 0)

        convert_tab_layout.setRowStretch(2, 1)
        self.convert_tab.setLayout(convert_tab_layout)

        self.addTab(self.convert_tab, "Conversion")

    def set_plot_transfer_function(self):
        print("Plotting transfer function...")
        self.current_plot_combobox.setCurrentIndex(1)

    def set_plot_spectrum(self):
        self.current_plot_combobox.setCurrentIndex(0)

    def on_current_plot_changed(self, current_plot):
        if current_plot == 'Frequency spectrum':
            self.sig_plot_frequency_spectrum.emit()
        elif current_plot == 'Transfer function':
            self.sig_plot_transfer_function.emit()



import sys
from PyQt5.QtCore import Qt, pyqtSignal
from PyQt5.QtWidgets import (QApplication, QWidget, QGridLayout, QTableWidget,
                             QDoubleSpinBox, QPushButton, QGroupBox,
                             QVBoxLayout, QHBoxLayout, QLabel, QComboBox,
                             QTreeWidget, QTreeWidgetItem,
                             QSpinBox, QRadioButton)


class CircleFitTree(QGroupBox):
    """
    The tree displaying the Circle Fit results and the parameters used for
    automatically fitting the circle.

    Attributes
    ----------
    sig_selected_peak_changed : pyqtSignal(int)
        The signal emitted when the selected peak is changed.
    """

    sig_selected_peak_changed = pyqtSignal(int)

    def __init__(self, parent=None, channelset=None):
        super().__init__("Results", parent)

        # TODO
        if channelset is None:
            self.num_channels = 5
        else:
            self.num_channels = len(channelset)

        self.num_peaks = 0

        self.init_ui()

    def init_ui(self):
        # # Tree for values
        self.tree = QTreeWidget()
        self.tree.setHeaderLabels(["Name", "Frequency", "Damping ratio",
                                   "Amplitude", "Phase", "Select"])
        self.tree.setStyleSheet("QTreeWidget::item:has-children { background-color : palette(mid);}")

        # # Tree for autofit parameters
        self.autofit_tree = QTreeWidget()
        self.autofit_tree.setHeaderLabels(["Name", "Frequency", "Damping ratio",
                                           "Amplitude", "Phase", "Select"])
        self.autofit_tree.hide()
        self.autofit_tree.setStyleSheet("QTreeWidget::item:has-children { background-color : palette(mid);}")

        # Connect the two trees together, so the views look identical
        self.autofit_tree.itemCollapsed.connect(self.on_item_collapsed)
        self.autofit_tree.itemExpanded.connect(self.on_item_expanded)
        self.tree.itemCollapsed.connect(self.on_item_collapsed)
        self.tree.itemExpanded.connect(self.on_item_expanded)

        # # Controls
        self.add_peak_btn = QPushButton("Add new peak", self)
        self.add_peak_btn.clicked.connect(self.add_peak)

        self.delete_selected_btn = QPushButton("Delete selected", self)
        self.delete_selected_btn.clicked.connect(self.delete_selected)

        self.view_autofit_btn = QPushButton("View autofit parameters", self)
        self.view_autofit_btn.clicked.connect(self.toggle_view)

        self.reset_to_autofit_btn = QPushButton("Reset all to auto", self)
        self.reset_to_autofit_btn.clicked.connect(self.reset_to_autofit)
        self.reset_to_autofit_btn.hide()

        controls1 = QGridLayout()
        controls1.addWidget(self.add_peak_btn, 0, 0)
        controls1.setColumnStretch(1, 1)
        controls1.addWidget(self.delete_selected_btn, 0, 2)

        controls2 = QGridLayout()
        controls2.setColumnStretch(0, 1)
        controls2.addWidget(self.reset_to_autofit_btn, 0, 1)
        controls2.addWidget(self.view_autofit_btn, 0, 2)

        # # Layout
        layout = QGridLayout()
        layout.addLayout(controls2, 0, 0)
        layout.addWidget(self.tree, 1, 0)
        layout.addWidget(self.autofit_tree, 1, 0)
        layout.addLayout(controls1, 2, 0)

        self.setLayout(layout)

    def add_peak(self):
        """Add a top-level item for a new peak to the tree, with children for
        each channel."""
        # Create the parent item for the peak
        peak_item = QTreeWidgetItem(self.tree,
                                    ["Peak {}".format(self.num_peaks),
                                     "", "", "", "", ""])
        autofit_peak_item = QTreeWidgetItem(self.autofit_tree,
                                            ["Peak {}".format(self.num_peaks),
                                             "", "", "", "", ""])

        # Put a radio button in column 5 in the tree
        radio_btn = QRadioButton()
        radio_btn.toggled.connect(self.on_selected_peak_changed)
        self.tree.setItemWidget(peak_item, 5, radio_btn)
        radio_btn.setChecked(True)

        for col in [1, 2, 3, 4]:
            # Put spinboxes in the tree
            spinbox = QDoubleSpinBox()
            self.tree.setItemWidget(peak_item, col, spinbox)
            # Put comboboxes in the autofit tree
            combobox = QComboBox()
            combobox.addItems(["Auto", "Manual"])
            combobox.currentIndexChanged.connect(self.update_peak_average)
            self.autofit_tree.setItemWidget(autofit_peak_item, col, combobox)

        # Create the child items for each channel for this peak if there's
        # more than one channel
        if self.num_channels > 1:
            for i in range(self.num_channels):
                channel_item = QTreeWidgetItem(peak_item,
                                               ["Channel {}".format(i),
                                                "", "", "", "", ""])
                autofit_channel_item = QTreeWidgetItem(autofit_peak_item,
                                                       ["Channel {}".format(i),
                                                        "", "", "", "", ""])
                for col in [1, 2, 3, 4]:
                    # Put spinboxes in the tree
                    spinbox = QDoubleSpinBox()
                    spinbox.valueChanged.connect(self.update_peak_average)
                    self.tree.setItemWidget(channel_item, col, spinbox)
                    # Put comboboxes in the autofit tree
                    combobox = QComboBox()
                    combobox.addItems(["Auto", "Manual"])
                    combobox.currentIndexChanged.connect(self.update_peak_average)
                    self.autofit_tree.setItemWidget(autofit_channel_item,
                                                    col, combobox)

        # Register that we've added another peak
        self.num_peaks += 1

    def delete_selected(self):
        """Delete the item that is currently selected."""
        for i in range(self.tree.topLevelItemCount()):
            # If the radio button is checked
            peak_item = self.tree.topLevelItem(i)
            if peak_item is not None:
                if self.tree.itemWidget(peak_item, 5).isChecked():
                    # Delete this item
                    self.tree.takeTopLevelItem(i)
                    self.autofit_tree.takeTopLevelItem(i)
                    #self.num_peaks -= 1

    def toggle_view(self):
        if self.tree.isVisible():
            self.tree.hide()
            self.autofit_tree.show()
            self.reset_to_autofit_btn.show()
            self.view_autofit_btn.setText("View parameter values")
        else:
            self.tree.show()
            self.autofit_tree.hide()
            self.reset_to_autofit_btn.hide()
            self.view_autofit_btn.setText("View autofit parameters")

    def on_item_collapsed(self, item):
        index = self.sender().indexOfTopLevelItem(item)
        self.tree.collapseItem(self.tree.topLevelItem(index))
        self.autofit_tree.collapseItem(self.autofit_tree.topLevelItem(index))

    def on_item_expanded(self, item):
        index = self.sender().indexOfTopLevelItem(item)
        self.tree.expandItem(self.tree.topLevelItem(index))
        self.autofit_tree.expandItem(self.autofit_tree.topLevelItem(index))

    def on_selected_peak_changed(self, checked):
        for i in range(self.tree.topLevelItemCount()):
            # If the radio button in this row is the sender
            peak_item = self.tree.topLevelItem(i)
            if peak_item is not None:
                if self.tree.itemWidget(peak_item, 5) == self.sender():
                    if checked:
                        print("Selected: " + str(i))
                        self.sig_selected_peak_changed.emit(i)

    def reset_to_autofit(self):
        """Reset all parameters to be automatically adjusted."""
        for peak_number in range(self.num_peaks):
            peak_item = self.autofit_tree.topLevelItem(peak_number)

            for col in [1, 2, 3, 4]:
                self.autofit_tree.itemWidget(peak_item, col).setCurrentIndex(0)

            if self.num_channels > 1:
                for channel_number in range(self.num_channels):
                    channel_item = peak_item.child(channel_number)
                    for col in [1, 2, 3, 4]:
                        self.autofit_tree.itemWidget(channel_item, col).setCurrentIndex(0)

    def update_peak_average(self):
        """Set the parameter values displayed for the peak to the average of
        all the channel values for each parameter."""
        for peak_number in range(self.num_peaks):
            # Get the peak item
            peak_item = self.tree.topLevelItem(peak_number)
            autofit_item = self.autofit_tree.topLevelItem(peak_number)

            if peak_item is not None:
                # Find the average values of all the channels
                avg_frequency = 0
                avg_damping = 0
                avg_amplitude = 0
                avg_phase = 0

                for channel_number in range(self.num_channels):
                    avg_frequency += self.get_frequency(peak_number, channel_number)
                    avg_damping += self.get_damping(peak_number, channel_number)
                    avg_amplitude += self.get_amplitude(peak_number, channel_number)
                    avg_phase += self.get_phase(peak_number, channel_number)

                avg_frequency /= self.num_channels
                avg_damping /= self.num_channels
                avg_amplitude /= self.num_channels
                avg_phase /= self.num_channels

                # Set the peak item to display the averages
                if self.autofit_tree.itemWidget(autofit_item, 1).currentText() == "Auto":
                    self.tree.itemWidget(peak_item, 1).setValue(avg_frequency)
                if self.autofit_tree.itemWidget(autofit_item, 2).currentText() == "Auto":
                    self.tree.itemWidget(peak_item, 2).setValue(avg_damping)
                if self.autofit_tree.itemWidget(autofit_item, 3).currentText() == "Auto":
                    self.tree.itemWidget(peak_item, 3).setValue(avg_amplitude)
                if self.autofit_tree.itemWidget(autofit_item, 4).currentText() == "Auto":
                    self.tree.itemWidget(peak_item, 4).setValue(avg_phase)

    def get_frequency(self, peak_number, channel_number=None):
        """Return the resonant frequency of the peak given by
        *peak_number*. If *channel_number* is given, return the resonant
        frequency of the given peak in the given channel."""
        peak_item = self.tree.topLevelItem(peak_number)
        if peak_item is not None:
            if channel_number is None:
                spinbox = self.tree.itemWidget(peak_item, 1)
                return spinbox.value()
            else:
                channel_item = peak_item.child(channel_number)
                spinbox = self.tree.itemWidget(channel_item, 1)
                return spinbox.value()
        else:
            return 0

    def get_damping(self, peak_number, channel_number=None):
        """Return the damping ratio of the peak given by *peak_number*. If
        *channel_number* is given, return the damping ratio of the given peak
        in the given channel."""
        peak_item = self.tree.topLevelItem(peak_number)
        if peak_item is not None:
            if channel_number is None:
                spinbox = self.tree.itemWidget(peak_item, 2)
                return spinbox.value()
            else:
                channel_item = peak_item.child(channel_number)
                spinbox = self.tree.itemWidget(channel_item, 2)
                return spinbox.value()
        else:
            return 0

    def get_amplitude(self, peak_number, channel_number=None):
        """Return the amplitude of the peak given by *peak_number*. If
        *channel_number* is given, return the amplitude of the given peak in
        the given channel."""
        peak_item = self.tree.topLevelItem(peak_number)
        if peak_item is not None:
            if channel_number is None:
                spinbox = self.tree.itemWidget(peak_item, 3)
                return spinbox.value()
            else:
                channel_item = peak_item.child(channel_number)
                spinbox = self.tree.itemWidget(channel_item, 3)
                return spinbox.value()
        else:
            return 0

    def get_phase(self, peak_number, channel_number=None):
        """Return the phase of the peak given by *peak_number*. If
        *channel_number* is given, return the phase of the given peak in the
        given channel."""
        peak_item = self.tree.topLevelItem(peak_number)
        if peak_item is not None:
            if channel_number is None:
                spinbox = self.tree.itemWidget(peak_item, 4)
                return spinbox.value()
            else:
                channel_item = peak_item.child(channel_number)
                spinbox = self.tree.itemWidget(channel_item, 4)
                return spinbox.value()
        else:
            return 0

    def get_parameter_values(self, peak_number, channel_number=None):
        """Returns a dict of the parameter values of a peak. If
        *channel_number* is given, return a dict of the parameter values of the
        peak for the given channel."""
        if channel_number is None:
            return {"frequency": self.get_frequency(peak_number),
                    "damping": self.get_damping(peak_number),
                    "amplitude": self.get_amplitude(peak_number),
                    "phase": self.get_phase(peak_number)}
        else:
            return \
                {"frequency": self.get_frequency(peak_number, channel_number),
                 "damping": self.get_damping(peak_number, channel_number),
                 "amplitude": self.get_amplitude(peak_number, channel_number),
                 "phase": self.get_phase(peak_number, channel_number)}

    def set_frequency(self, peak_number, channel_number=None, value=0):
        """Set the frequency of a given peak to *value*. If *channel_number* is
        given, set the frequency of the given peak in the given channel."""
        # Get the top level items
        peak_item = self.tree.topLevelItem(peak_number)
        autofit_peak_item = self.autofit_tree.topLevelItem(peak_number)
        # Check they exist
        if peak_item is not None:
            # If they have children
            if channel_number is not None:
                # Get the channel items
                channel_item = peak_item.child(channel_number)
                autofit_channel_item = autofit_peak_item.child(channel_number)
                # Set the value only if the combobox is set to autofit
                combobox = \
                    self.autofit_tree.itemWidget(autofit_channel_item, 1)
                if combobox.currentText() == "Auto":
                    spinbox = self.tree.itemWidget(channel_item, 1)
                    spinbox.setValue(value)
            else:
                # Set the value only if the combobox is set to autofit
                combobox = self.autofit_tree.itemWidget(autofit_peak_item, 1)
                if combobox.currentText() == "Auto":
                    spinbox = self.tree.itemWidget(peak_item, 1)
                    spinbox.setValue(value)

    def set_damping(self, peak_number, channel_number=None, value=0):
        """Set the damping ratio of a given peak to *value*. If *channel_number* is
        given, set the damping ratio of the given peak in the given channel."""
        # Get the top level items
        peak_item = self.tree.topLevelItem(peak_number)
        autofit_peak_item = self.autofit_tree.topLevelItem(peak_number)
        # Check they exist
        if peak_item is not None:
            # If they have children
            if channel_number is not None:
                # Get the channel items
                channel_item = peak_item.child(channel_number)
                autofit_channel_item = autofit_peak_item.child(channel_number)
                # Set the value only if the combobox is set to autofit
                combobox = \
                    self.autofit_tree.itemWidget(autofit_channel_item, 2)
                if combobox.currentText() == "Auto":
                    spinbox = self.tree.itemWidget(channel_item, 2)
                    spinbox.setValue(value)
            else:
                # Set the value only if the combobox is set to autofit
                combobox = self.autofit_tree.itemWidget(autofit_peak_item, 2)
                if combobox.currentText() == "Auto":
                    spinbox = self.tree.itemWidget(peak_item, 2)
                    spinbox.setValue(value)

    def set_amplitude(self, peak_number, channel_number=None, value=0):
        """Set the amplitude of a given peak to *value*. If *channel_number* is
        given, set the amplitude of the given peak in the given channel."""
        # Get the top level items
        peak_item = self.tree.topLevelItem(peak_number)
        autofit_peak_item = self.autofit_tree.topLevelItem(peak_number)
        # Check they exist
        if peak_item is not None:
            # If they have children
            if channel_number is not None:
                # Get the channel items
                channel_item = peak_item.child(channel_number)
                autofit_channel_item = autofit_peak_item.child(channel_number)
                # Set the value only if the combobox is set to autofit
                combobox = \
                    self.autofit_tree.itemWidget(autofit_channel_item, 3)
                if combobox.currentText() == "Auto":
                    spinbox = self.tree.itemWidget(channel_item, 3)
                    spinbox.setValue(value)
            else:
                # Set the value only if the combobox is set to autofit
                combobox = self.autofit_tree.itemWidget(autofit_peak_item, 3)
                if combobox.currentText() == "Auto":
                    spinbox = self.tree.itemWidget(peak_item, 3)
                    spinbox.setValue(value)

    def set_phase(self, peak_number, channel_number=None, value=0):
        """Set the phase of a given peak to *value*. If *channel_number* is
        given, set the phase of the given peak in the given channel."""
        # Get the top level items
        peak_item = self.tree.topLevelItem(peak_number)
        autofit_peak_item = self.autofit_tree.topLevelItem(peak_number)
        # Check they exist
        if peak_item is not None:
            # If they have children
            if channel_number is not None:
                # Get the channel items
                channel_item = peak_item.child(channel_number)
                autofit_channel_item = autofit_peak_item.child(channel_number)
                # Set the value only if the combobox is set to autofit
                combobox = \
                    self.autofit_tree.itemWidget(autofit_channel_item, 4)
                if combobox.currentText() == "Auto":
                    spinbox = self.tree.itemWidget(channel_item, 4)
                    spinbox.setValue(value)
            else:
                # Set the value only if the combobox is set to autofit
                combobox = self.autofit_tree.itemWidget(autofit_peak_item, 4)
                if combobox.currentText() == "Auto":
                    spinbox = self.tree.itemWidget(peak_item, 4)
                    spinbox.setValue(value)

    def set_parameter_values(self, peak_number, channel_number=None,
                             channel_values={}):
        """Set the parameter values for a given peak to those given in
        *channel_values*. If *channel_number* is given, set the values of the
        given peak in the given channel."""
        if "frequency" in channel_values.keys():
            self.set_frequency(peak_number, channel_number,
                               channel_values["frequency"])
        if "damping" in channel_values.keys():
            self.set_damping(peak_number, channel_number,
                             channel_values["damping"])
        if "amplitude" in channel_values.keys():
            self.set_amplitude(peak_number, channel_number,
                               channel_values["amplitude"])
        if "phase" in channel_values.keys():
            self.set_phase(peak_number, channel_number,
                           channel_values["phase"])

if __name__ == '__main__':
    app=0
    app=QApplication(sys.argv)
    w = CircleFitTree()
    w.show()
    sys.exit(app.exec_())


from cued_datalogger.api.pyqtgraph_extensions import InteractivePlotWidget
from cued_datalogger.api.toolbox import Toolbox

from PyQt5.QtWidgets import QWidget, QGridLayout, QPushButton
from PyQt5.QtCore import pyqtSignal


class TimeDomainWidget(InteractivePlotWidget):
    """
    The TimeDomainWidget is the main display widget for everything in
    the time domain.
    """
    def __init__(self, parent=None):
        super().__init__(parent)
        self.channels = []

    def set_selected_channels(self, selected_channels):
        """Update which channels are plotted"""
        # If no channel list is given
        if not selected_channels:
            self.channels = []
        else:
            self.channels = selected_channels
        self.update_plot()

    def update_plot(self):
        self.clear()
        for channel in self.channels:
            if channel.is_dataset("time_series"):
                self.plot(channel.data("time"),
                          channel.data("time_series"),
                          pen=channel.colour)


class TimeToolbox(Toolbox):
    """
    Toolbox containing the Time Domain controls.

    Attributes
    ----------
    sig_convert_to_sonogram : pyqtSignal
        Signal emitted when the 'Convert to sonogram' button is clicked.
    sig_convert_to_fft : pyqtSignal
        Signal emitted when the 'Convert to frequency spectrum' button is
        clicked.

    """
    sig_convert_to_sonogram = pyqtSignal()
    sig_convert_to_fft = pyqtSignal()

    def __init__(self, parent=None):
        super().__init__(parent=parent)
        self.parent = parent
        self.init_ui()

    def init_ui(self):
        self.convert_tab = QWidget(self)
        convert_tab_layout = QGridLayout()

        self.fft_btn = QPushButton("Convert to frequency spectrum")
        self.fft_btn.clicked.connect(self.sig_convert_to_fft.emit)
        convert_tab_layout.addWidget(self.fft_btn, 0, 0)

        self.sonogram_btn = QPushButton("Convert to sonogram")
        self.sonogram_btn.clicked.connect(self.sig_convert_to_sonogram.emit)
        convert_tab_layout.addWidget(self.sonogram_btn, 1, 0)

        convert_tab_layout.setRowStretch(2, 1)

        self.convert_tab.setLayout(convert_tab_layout)

        self.addTab(self.convert_tab, "Conversion")



# -*- coding: utf-8 -*-
"""
Created on Wed Jul 19 11:28:37 2017

@author: eyt21

PyQt signals for the Recorder Classes
"""

from PyQt5.QtCore import QObject, pyqtSignal

class RecEmitter(QObject):
    """
    PyQt signals for the Recorder Classes
    
    Attributes
    ----------
    recorddone: pyqtsignal
        Emits when recording is done
    triggered: pyqtsignal
        Emits when trigger threshold is reached
    newdata: pyqtsignal
        Emits when new data is received (not used)
    """
    recorddone = pyqtSignal()
    triggered = pyqtSignal()
    newdata = pyqtSignal()